import re
from typing import Any, Dict, Optional, Type, Union, ValuesView

from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    ArrayType,
    BooleanType,
    BytesType,
    DateType,
    EnumType,
    MapType,
    NullType,
    NumberType,
    RecordType,
    StringType,
    TimeType,
    UnionType,
)

DATAHUB_FIELD_TYPE = Union[
    ArrayType,
    BooleanType,
    BytesType,
    DateType,
    EnumType,
    MapType,
    NullType,
    NumberType,
    RecordType,
    StringType,
    TimeType,
    UnionType,
]


# These can be obtained by running `select format_type(oid, null),* from pg_type;`
# We've omitted the types without a meaningful DataHub type (e.g. postgres-specific types, index vectors, etc.)
# (run `\copy (select format_type(oid, null),* from pg_type) to 'pg_type.csv' csv header;` to get a CSV)
# We map from format_type since this is what dbt uses.
# See https://github.com/fishtown-analytics/dbt/blob/master/plugins/postgres/dbt/include/postgres/macros/catalog.sql#L22
# See https://www.npgsql.org/dev/types.html for helpful type annotations
POSTGRES_TYPES_MAP: Dict[str, Any] = {
    "boolean": BooleanType,
    "bytea": BytesType,
    '"char"': StringType,
    "name": None,  # for system identifiers
    "bigint": NumberType,
    "smallint": NumberType,
    "int2vector": NumberType,  # for indexing
    "integer": NumberType,
    "regproc": None,  # object identifier
    "text": StringType,
    "oid": None,  # object identifier
    "tid": None,  # object identifier
    "xid": None,  # object identifier
    "cid": None,  # object identifier
    "oidvector": None,  # object identifier
    "json": RecordType,
    "xml": RecordType,
    "xid8": None,  # object identifier
    "point": None,  # 2D point
    "lseg": None,  # line segment
    "path": None,  # path of points
    "box": None,  # a pair of corner points
    "polygon": None,  # closed set of points
    "line": None,  # infinite line
    "real": NumberType,
    "double precision": NumberType,
    "unknown": None,
    "circle": None,  # circle with center and radius
    "money": NumberType,
    "macaddr": None,  # MAC address
    "inet": None,  # IPv4 or IPv6 host address
    "cidr": None,  # IPv4 or IPv6 network specification
    "macaddr8": None,  # MAC address
    "aclitem": None,  # system info
    "character": StringType,
    "character varying": StringType,
    "date": DateType,
    "time without time zone": TimeType,
    "timestamp without time zone": TimeType,
    "timestamp with time zone": TimeType,
    "interval": None,
    "time with time zone": TimeType,
    "bit": BytesType,
    "bit varying": BytesType,
    "numeric": NumberType,
    "refcursor": None,
    "regprocedure": None,
    "regoper": None,
    "regoperator": None,
    "regclass": None,
    "regcollation": None,
    "regtype": None,
    "regrole": None,
    "regnamespace": None,
    "super": NullType,
    "uuid": StringType,
    "pg_lsn": None,
    "tsvector": None,  # text search vector
    "gtsvector": None,  # GiST for tsvector. Probably internal type.
    "tsquery": None,  # text search query tree
    "regconfig": None,
    "regdictionary": None,
    "jsonb": BytesType,
    "jsonpath": None,  # path to property in a JSON doc
    "txid_snapshot": None,
    "pg_snapshot": None,
    "int4range": None,  # don't have support for ranges yet
    "numrange": None,
    "tsrange": None,
    "tstzrange": None,
    "daterange": None,
    "int8range": None,
    "record": RecordType,
    "record[]": ArrayType,
    "cstring": None,
    '"any"': UnionType,
    "anyarray": ArrayType,
    "void": NullType,
    "trigger": None,
    "event_trigger": None,
    "language_handler": None,
    "internal": None,
    "anyelement": None,
    "anynonarray": None,
    "anyenum": EnumType,
    "fdw_handler": None,
    "index_am_handler": None,
    "tsm_handler": None,
    "table_am_handler": None,
    "anyrange": None,
    "anycompatible": None,
    "anycompatiblearray": None,
    "anycompatiblenonarray": None,
    "anycompatiblerange": None,
    "boolean[]": ArrayType,
    "bytea[]": ArrayType,
    '"char"[]': ArrayType,
    "name[]": ArrayType,
    "bigint[]": ArrayType,
    "smallint[]": ArrayType,
    "int2vector[]": ArrayType,
    "integer[]": ArrayType,
    "regproc[]": ArrayType,
    "text[]": ArrayType,
    "oid[]": ArrayType,
    "tid[]": ArrayType,
    "xid[]": ArrayType,
    "cid[]": ArrayType,
    "oidvector[]": ArrayType,
    "json[]": ArrayType,
    "xml[]": ArrayType,
    "xid8[]": ValuesView,
    "point[]": ArrayType,
    "lseg[]": ArrayType,
    "path[]": ArrayType,
    "box[]": ArrayType,
    "polygon[]": ArrayType,
    "line[]": ArrayType,
    "real[]": ArrayType,
    "double precision[]": ArrayType,
    "circle[]": ArrayType,
    "money[]": ArrayType,
    "macaddr[]": ArrayType,
    "inet[]": ArrayType,
    "cidr[]": ArrayType,
    "macaddr8[]": ArrayType,
    "aclitem[]": ArrayType,
    "character[]": ArrayType,
    "character varying[]": ArrayType,
    "date[]": ArrayType,
    "time without time zone[]": ArrayType,
    "timestamp without time zone[]": ArrayType,
    "timestamp with time zone[]": ArrayType,
    "interval[]": ArrayType,
    "time with time zone[]": ArrayType,
    "bit[]": ArrayType,
    "bit varying[]": ArrayType,
    "numeric[]": ArrayType,
    "refcursor[]": ArrayType,
    "regprocedure[]": ArrayType,
    "regoper[]": ArrayType,
    "regoperator[]": ArrayType,
    "regclass[]": ArrayType,
    "regcollation[]": ArrayType,
    "regtype[]": ArrayType,
    "regrole[]": ArrayType,
    "regnamespace[]": ArrayType,
    "uuid[]": ArrayType,
    "pg_lsn[]": ArrayType,
    "tsvector[]": ArrayType,
    "gtsvector[]": ArrayType,
    "tsquery[]": ArrayType,
    "regconfig[]": ArrayType,
    "regdictionary[]": ArrayType,
    "jsonb[]": ArrayType,
    "jsonpath[]": ArrayType,
    "txid_snapshot[]": ArrayType,
    "pg_snapshot[]": ArrayType,
    "int4range[]": ArrayType,
    "numrange[]": ArrayType,
    "tsrange[]": ArrayType,
    "tstzrange[]": ArrayType,
    "daterange[]": ArrayType,
    "int8range[]": ArrayType,
    "cstring[]": ArrayType,
}

# Postgres types with modifiers (identifiable by non-empty typmodin and typmodout columns)
POSTGRES_MODIFIED_TYPES = {
    "character varying",
    "character varying[]",
    "bit varying",
    "bit varying[]",
    "time with time zone",
    "time with time zone[]",
    "time without time zone",
    "time without time zone[]",
    "timestamp with time zone",
    "timestamp with time zone[]",
    "timestamp without time zone",
    "timestamp without time zone[]",
    "numeric",
    "numeric[]",
    "interval",
    "interval[]",
    "character",
    "character[]",
    "bit",
    "bit[]",
}


def resolve_postgres_modified_type(type_string: str) -> Any:
    if type_string.endswith("[]"):
        return ArrayType

    for modified_type_base in POSTGRES_MODIFIED_TYPES:
        if re.match(rf"{re.escape(modified_type_base)}\([0-9,]+\)", type_string):
            return POSTGRES_TYPES_MAP[modified_type_base]

    return None


def resolve_trino_modified_type(type_string: str) -> Any:
    # for cases like timestamp(3), decimal(10,0), row(...)
    match = re.match(r"([a-zA-Z]+)\(.+\)", type_string)
    if match:
        modified_type_base: str = match.group(1)
        return TRINO_SQL_TYPES_MAP[modified_type_base]
    return TRINO_SQL_TYPES_MAP[type_string]


def resolve_athena_modified_type(type_string: str) -> Any:
    # for cases like struct<...>, array<...>, map<...>
    match_complex = re.match(r"([a-zA-Z]+)<.+>", type_string)
    # for cases like timestamp(3), decimal(10,0)
    match_simple = re.match(r"([a-zA-Z]+)\(.+\)", type_string)

    modified_type_base = ""
    if match_complex:
        modified_type_base = match_complex.group(1)
    elif match_simple:
        modified_type_base = match_simple.group(1)
    if modified_type_base:
        return ATHENA_SQL_TYPES_MAP[modified_type_base]
    return ATHENA_SQL_TYPES_MAP[type_string]


def resolve_vertica_modified_type(type_string: str) -> Any:
    # for cases like timestamp(3), decimal(10,0)
    match = re.match(r"([a-zA-Z ]+)\(.+\)", type_string)
    if match:
        modified_type_base: str = match.group(1)
        return VERTICA_SQL_TYPES_MAP[modified_type_base]
    return VERTICA_SQL_TYPES_MAP[type_string]


SNOWFLAKE_TYPES_MAP: Dict[str, Any] = {
    "NUMBER": NumberType,
    "DECIMAL": NumberType,
    "NUMERIC": NumberType,
    "INT": NumberType,
    "INTEGER": NumberType,
    "BIGINT": NumberType,
    "SMALLINT": NumberType,
    "TINYINT": NumberType,
    "BYTEINT": NumberType,
    "FLOAT": NumberType,
    "FLOAT4": NumberType,
    "FLOAT8": NumberType,
    "DOUBLE": NumberType,
    "DOUBLE PRECISION": NumberType,
    "REAL": NumberType,
    "VARCHAR": StringType,
    "CHARACTER VARYING": StringType,
    "CHAR": StringType,
    "CHARACTER": StringType,
    "STRING": StringType,
    "TEXT": StringType,
    "BINARY": BytesType,
    "VARBINARY": BytesType,
    "BOOLEAN": BooleanType,
    "DATE": DateType,
    "DATETIME": DateType,
    "TIME": TimeType,
    "TIMESTAMP": TimeType,
    "TIMESTAMP_LTZ": TimeType,
    "TIMESTAMP_NTZ": TimeType,
    "TIMESTAMP_TZ": TimeType,
    "VARIANT": RecordType,
    "OBJECT": RecordType,
    "ARRAY": ArrayType,
    "GEOGRAPHY": None,
}


def resolve_snowflake_modified_type(type_string: str) -> Any:
    # Match types with precision and scale, e.g., 'DECIMAL(38,0)' or TIME(3)
    match = re.match(r"([a-z A-Z_]+)\(\d+(,(\s+)?\d+)?\)", type_string)
    if match:
        modified_type_base = match.group(1)  # Extract the base type
        return SNOWFLAKE_TYPES_MAP.get(modified_type_base)

    # Fallback for types without precision/scale
    return SNOWFLAKE_TYPES_MAP.get(type_string)


# see https://github.com/googleapis/python-bigquery-sqlalchemy/blob/main/sqlalchemy_bigquery/_types.py#L32
BIGQUERY_TYPES_MAP: Dict[str, Any] = {
    "STRING": StringType,
    "BOOL": BooleanType,
    "BOOLEAN": BooleanType,
    "INT64": NumberType,
    "INTEGER": NumberType,
    "FLOAT64": NumberType,
    "FLOAT": NumberType,
    "TIMESTAMP": TimeType,
    "DATETIME": DateType,
    "DATE": DateType,
    "BYTES": BytesType,
    "TIME": TimeType,
    "RECORD": RecordType,
    "NUMERIC": NumberType,
    "BIGNUMERIC": NumberType,
    "GEOGRAPHY": None,
}

# see https://spark.apache.org/docs/latest/sql-ref-datatypes.html
SPARK_SQL_TYPES_MAP: Dict[str, Any] = {
    "boolean": BooleanType,
    "byte": NumberType,
    "tinyint": NumberType,
    "short": NumberType,
    "smallint": NumberType,
    "int": NumberType,
    "integer": NumberType,
    "long": NumberType,
    "bigint": NumberType,
    "float": NumberType,
    "real": NumberType,
    "double": NumberType,
    "date": DateType,
    "timestamp": TimeType,
    "string": StringType,
    "binary": BytesType,
    "decimal": NumberType,
    "dec": NumberType,
    "numeric": NumberType,
    "array": ArrayType,
    "struct": RecordType,
    "map": RecordType,
}

# https://trino.io/docs/current/language/types.html
# https://github.com/trinodb/trino-python-client/blob/master/trino/sqlalchemy/datatype.py#L75
TRINO_SQL_TYPES_MAP: Dict[str, Any] = {
    "boolean": BooleanType,
    "tinyint": NumberType,
    "smallint": NumberType,
    "int": NumberType,
    "integer": NumberType,
    "bigint": NumberType,
    "real": NumberType,
    "double": NumberType,
    "decimal": NumberType,
    "varchar": StringType,
    "char": StringType,
    "varbinary": BytesType,
    "date": DateType,
    "time": TimeType,
    "timestamp": TimeType,
    "row": RecordType,
    "map": MapType,
    "array": ArrayType,
    "json": RecordType,
}

# https://docs.aws.amazon.com/athena/latest/ug/data-types.html
# https://github.com/dbt-athena/dbt-athena/tree/main
ATHENA_SQL_TYPES_MAP: Dict[str, Any] = {
    "boolean": BooleanType,
    "tinyint": NumberType,
    "smallint": NumberType,
    "int": NumberType,
    "integer": NumberType,
    "bigint": NumberType,
    "float": NumberType,
    "double": NumberType,
    "decimal": NumberType,
    "varchar": StringType,
    "char": StringType,
    "binary": BytesType,
    "date": DateType,
    "timestamp": TimeType,
    "struct": RecordType,
    "map": MapType,
    "array": ArrayType,
    "row": RecordType,
}

# https://www.vertica.com/docs/11.1.x/HTML/Content/Authoring/SQLReferenceManual/DataTypes/SQLDataTypes.htm
VERTICA_SQL_TYPES_MAP: Dict[str, Any] = {
    "binary": BytesType,
    "varbinary": BytesType,
    "long varbinary": BytesType,
    "bytea": BytesType,
    "raw": BytesType,
    "boolean": BooleanType,
    "char": StringType,
    "varchar": StringType,
    "long varchar": StringType,
    "date": DateType,
    "time": TimeType,
    "datetime": TimeType,
    "smalldatetime": TimeType,
    "time with timezone": TimeType,
    "timestamp": TimeType,
    "timestamptz": TimeType,
    "timestamp with timezone": TimeType,
    "interval": TimeType,
    "interval hour to second": TimeType,
    "double precision": NumberType,
    "float": NumberType,
    "float8": NumberType,
    "real": NumberType,
    "integer": NumberType,
    "int": NumberType,
    "bigint": NumberType,
    "int8": NumberType,
    "smallint": NumberType,
    "tinyint": NumberType,
    "decimal": NumberType,
    "numeric": NumberType,
    "number": NumberType,
    "money": NumberType,
    "geometry": None,
    "geography": None,
    "uuid": StringType,
}

# Neo4j property types mapping
# https://neo4j.com/docs/cypher-manual/current/values-and-types/property-structural-constructed/
NEO4J_TYPES_MAP: Dict[str, Any] = {
    "boolean": BooleanType,
    "date": DateType,
    "duration": TimeType,  # Neo4j duration represents a temporal amount
    "float": NumberType,
    "integer": NumberType,
    "list": ArrayType,
    "local_date_time": TimeType,
    "local_time": TimeType,
    "point": StringType,  # Neo4j point - spatial coordinate, represented as string
    "string": StringType,
    "zoned_date_time": TimeType,
    "zoned_time": TimeType,
    "node": StringType,  # Neo4j object type
    "relationship": StringType,  # Neo4j object type
}

# Google Dataplex type mapping
# https://cloud.google.com/dataplex/docs/reference/rest/v1/projects.locations.lakes.zones.entities#Schema.Type
DATAPLEX_TYPES_MAP: Dict[str, Any] = {
    "BOOL": BooleanType,
    "BYTE": BytesType,
    "INT16": NumberType,
    "INT32": NumberType,
    "INT64": NumberType,
    "FLOAT": NumberType,
    "DOUBLE": NumberType,
    "DECIMAL": NumberType,
    "STRING": StringType,
    "BINARY": BytesType,
    "TIMESTAMP": TimeType,
    "DATE": DateType,
    "TIME": TimeType,
    "RECORD": RecordType,
    "NULL": NullType,
}


_merged_mapping = {
    "boolean": BooleanType,
    "date": DateType,
    "time": TimeType,
    "numeric": NumberType,
    "text": StringType,
    "timestamp with time zone": DateType,
    "timestamp without time zone": DateType,
    "integer": NumberType,
    "float8": NumberType,
    "struct": RecordType,
    **POSTGRES_TYPES_MAP,
    **SNOWFLAKE_TYPES_MAP,
    **BIGQUERY_TYPES_MAP,
    **SPARK_SQL_TYPES_MAP,
    **TRINO_SQL_TYPES_MAP,
    **ATHENA_SQL_TYPES_MAP,
    **VERTICA_SQL_TYPES_MAP,
    **NEO4J_TYPES_MAP,
    **DATAPLEX_TYPES_MAP,
}


def resolve_sql_type(
    column_type: Optional[str],
    platform: Optional[str] = None,
) -> Optional[DATAHUB_FIELD_TYPE]:
    # In theory, we should use the platform-specific mapping where available.
    # However, the types don't ever conflict, so the merged mapping is fine.
    # Wrong assumption - there ARE conflicts as the test_type_conflicts_across_platforms in test_sql_types.py shows.
    # TODO: revisit this and make platform-specific mappings work.
    TypeClass: Optional[Type[DATAHUB_FIELD_TYPE]] = (
        _merged_mapping.get(column_type) if column_type else None
    )

    if TypeClass is None and column_type:
        # resolve a modified type
        if platform == "trino":
            TypeClass = resolve_trino_modified_type(column_type)
        elif platform == "athena":
            TypeClass = resolve_athena_modified_type(column_type)
        elif platform == "postgres" or platform == "redshift":
            # Redshift uses a variant of Postgres, so we can use the same logic.
            TypeClass = resolve_postgres_modified_type(column_type)
        elif platform == "vertica":
            TypeClass = resolve_vertica_modified_type(column_type)
        elif platform == "snowflake":
            # Snowflake types are uppercase, so we check that.
            TypeClass = resolve_snowflake_modified_type(column_type.upper())

    if TypeClass:
        return TypeClass()
    return None



"""Unified Azure authentication module for DataHub connectors.

This module provides a reusable authentication configuration that can be used
across all Azure connectors (ADF, Synapse, Fabric, etc.).

Supports multiple authentication methods:
- Service Principal (client_id + client_secret + tenant_id)
- Managed Identity (system-assigned or user-assigned)
- Azure CLI credentials (for local development)
- DefaultAzureCredential (auto-detects environment)
"""

from typing import Optional

from azure.core.credentials import TokenCredential
from azure.identity import (
    AzureCliCredential,
    ClientSecretCredential,
    DefaultAzureCredential,
    ManagedIdentityCredential,
)
from pydantic import Field, SecretStr, model_validator

from datahub.configuration import ConfigModel
from datahub.utilities.str_enum import StrEnum


class AzureAuthenticationMethod(StrEnum):
    """Supported Azure authentication methods.

    - DEFAULT: Uses DefaultAzureCredential which auto-detects credentials from
      environment variables, managed identity, Azure CLI, etc.
    - SERVICE_PRINCIPAL: Uses client ID, client secret, and tenant ID
    - MANAGED_IDENTITY: Uses Azure Managed Identity (system or user-assigned)
    - CLI: Uses Azure CLI credential (requires `az login`)
    """

    DEFAULT = "default"
    SERVICE_PRINCIPAL = "service_principal"
    MANAGED_IDENTITY = "managed_identity"
    CLI = "cli"


class AzureCredentialConfig(ConfigModel):
    """Unified Azure authentication configuration.

    This class provides a reusable authentication configuration that can be
    composed into any Azure connector's configuration. It supports multiple
    authentication methods and returns a TokenCredential that works with
    any Azure SDK client.

    Example usage in a connector config:
        class MyAzureConnectorConfig(ConfigModel):
            credential: AzureCredentialConfig = Field(
                default_factory=AzureCredentialConfig,
                description="Azure authentication configuration"
            )
            subscription_id: str = Field(...)
    """

    authentication_method: AzureAuthenticationMethod = Field(
        default=AzureAuthenticationMethod.DEFAULT,
        description=(
            "Authentication method to use. Options: "
            "'default' (auto-detects from environment), "
            "'service_principal' (client ID + secret + tenant), "
            "'managed_identity' (Azure Managed Identity), "
            "'cli' (Azure CLI credential). "
            "Recommended: Use 'default' which tries multiple methods automatically."
        ),
    )

    # Service Principal credentials (required when authentication_method = "service_principal")
    client_id: Optional[str] = Field(
        default=None,
        description=(
            "Azure Application (client) ID. Required for service_principal authentication. "
            "Find this in Azure Portal > App registrations > Your app > Overview."
        ),
    )
    client_secret: Optional[SecretStr] = Field(
        default=None,
        description=(
            "Azure client secret. Required for service_principal authentication. "
            "Create in Azure Portal > App registrations > Your app > Certificates & secrets."
        ),
    )
    tenant_id: Optional[str] = Field(
        default=None,
        description=(
            "Azure tenant (directory) ID. Required for service_principal authentication. "
            "Find this in Azure Portal > Microsoft Entra ID > Overview."
        ),
    )

    # Managed Identity options (optional, for user-assigned managed identity)
    managed_identity_client_id: Optional[str] = Field(
        default=None,
        description=(
            "Client ID for user-assigned managed identity. "
            "Leave empty to use system-assigned managed identity. "
            "Only used when authentication_method is 'managed_identity'."
        ),
    )

    # Additional options for DefaultAzureCredential
    exclude_cli_credential: bool = Field(
        default=False,
        description=(
            "When using 'default' authentication, exclude Azure CLI credential. "
            "Useful in production to avoid accidentally using developer credentials."
        ),
    )
    exclude_environment_credential: bool = Field(
        default=False,
        description=(
            "When using 'default' authentication, exclude environment variables. "
            "Environment variables checked: AZURE_CLIENT_ID, AZURE_CLIENT_SECRET, AZURE_TENANT_ID."
        ),
    )
    exclude_managed_identity_credential: bool = Field(
        default=False,
        description=(
            "When using 'default' authentication, exclude managed identity. "
            "Useful during local development when managed identity is not available."
        ),
    )

    def get_credential(self) -> TokenCredential:
        """Get Azure credential based on the configured authentication method.

        Returns:
            TokenCredential: An Azure credential object that can be used with
                any Azure SDK client (e.g., DataFactoryManagementClient).

        Raises:
            ValueError: If required credentials are missing for the chosen method.
        """
        if self.authentication_method == AzureAuthenticationMethod.SERVICE_PRINCIPAL:
            # Validate all required fields (also validated in validate_credentials())
            if not self.client_secret:
                raise ValueError(
                    "client_secret is required for service_principal authentication"
                )
            if not self.tenant_id:
                raise ValueError(
                    "tenant_id is required for service_principal authentication"
                )
            if not self.client_id:
                raise ValueError(
                    "client_id is required for service_principal authentication"
                )
            return ClientSecretCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret.get_secret_value(),
            )

        elif self.authentication_method == AzureAuthenticationMethod.MANAGED_IDENTITY:
            return ManagedIdentityCredential(client_id=self.managed_identity_client_id)

        elif self.authentication_method == AzureAuthenticationMethod.CLI:
            return AzureCliCredential()

        else:  # DEFAULT
            return DefaultAzureCredential(
                exclude_cli_credential=self.exclude_cli_credential,
                exclude_environment_credential=self.exclude_environment_credential,
                exclude_managed_identity_credential=self.exclude_managed_identity_credential,
            )

    @model_validator(mode="after")
    def validate_credentials(self) -> "AzureCredentialConfig":
        """Validate that required credentials are provided for the chosen method."""
        if self.authentication_method == AzureAuthenticationMethod.SERVICE_PRINCIPAL:
            missing = []
            if not self.client_id:
                missing.append("client_id")
            if not self.client_secret:
                missing.append("client_secret")
            if not self.tenant_id:
                missing.append("tenant_id")

            if missing:
                raise ValueError(
                    f"Service principal authentication requires: {', '.join(missing)}. "
                    f"These can be found in Azure Portal > App registrations."
                )

        return self



import contextlib
import functools
import platform
from typing import ContextManager

from stopit import ThreadingTimeout as _ThreadingTimeout, TimeoutException

__all__ = ["threading_timeout", "TimeoutException"]


@functools.lru_cache(maxsize=1)
def _is_cpython() -> bool:
    """Check if we're running on CPython."""
    return platform.python_implementation() == "CPython"


def threading_timeout(timeout: float) -> ContextManager[None]:
    """A timeout context manager that uses stopit's ThreadingTimeout underneath.

    This is only supported on CPython.
    That's because stopit.ThreadingTimeout uses a CPython-internal method to raise
    an exception (the timeout error) in another thread. See stopit.threadstop.async_raise.

    Reference: https://github.com/glenfant/stopit

    Args:
        timeout: The timeout in seconds. If <= 0, no timeout is applied.

    Raises:
        RuntimeError: If the timeout is not supported on the current Python implementation.
        TimeoutException: If the timeout is exceeded.
    """

    if timeout <= 0:
        return contextlib.nullcontext()

    if not _is_cpython():
        raise RuntimeError(
            f"Timeout is only supported on CPython, not {platform.python_implementation()}"
        )

    return _ThreadingTimeout(timeout, swallow_exc=False)



import dataclasses
import json
from typing import Dict, Iterable, Optional

from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.schema_classes import (
    ChangeTypeClass,
    DatasetPropertiesClass,
    GenericAspectClass,
    MetadataChangeProposalClass,
    OperationClass,
    TimeStampClass,
)
from datahub.specific.dataset import DatasetPatchBuilder
from datahub.utilities.urns.urn import guess_entity_type


@dataclasses.dataclass
class TimestampPair:
    last_modified_dataset_props: Optional[
        TimeStampClass
    ]  # last_modified of datasetProperties aspect
    last_updated_timestamp_dataset_props: Optional[
        int
    ]  # lastUpdatedTimestamp of the operation aspect


def try_aspect_from_metadata_change_proposal_class(
    wu: MetadataWorkUnit,
) -> Optional[DatasetPropertiesClass]:
    if (
        isinstance(wu.metadata, MetadataChangeProposalClass)
        and wu.metadata.aspectName == "datasetProperties"
        and wu.metadata.changeType == ChangeTypeClass.PATCH
        and isinstance(wu.metadata.aspect, GenericAspectClass)
    ):
        patch_dataset_properties = json.loads(wu.metadata.aspect.value)
        for operation in patch_dataset_properties:
            if operation.get("path") == "/lastModified":
                # Deserializing `lastModified` as the `auto_patch_last_modified` function relies on this property
                # to decide if a patch aspect for the datasetProperties aspect should be generated
                return DatasetPropertiesClass(
                    lastModified=TimeStampClass(time=operation["value"]["time"])
                )

    return None


def auto_patch_last_modified(
    stream: Iterable[MetadataWorkUnit],
) -> Iterable[MetadataWorkUnit]:
    """
    Generate a patch request for datasetProperties aspect in-case
        1. `lastModified` of datasetProperties is not set
        2. And there are operation aspects
    in this case set the `lastModified` of datasetProperties to max value of operation aspects `lastUpdatedTimestamp`.

    We need this functionality to support sort by `last modified` on UI.
    """
    candidate_dataset_for_patch: Dict[str, TimestampPair] = {}

    for wu in stream:
        if (
            guess_entity_type(wu.get_urn()) != "dataset"
        ):  # we are only processing datasets
            yield wu
            continue

        dataset_properties_aspect = wu.get_aspect_of_type(
            DatasetPropertiesClass
        ) or try_aspect_from_metadata_change_proposal_class(wu)
        dataset_operation_aspect = wu.get_aspect_of_type(OperationClass)

        timestamp_pair = candidate_dataset_for_patch.get(wu.get_urn())

        if timestamp_pair:
            # Update the timestamp_pair
            if dataset_properties_aspect and dataset_properties_aspect.lastModified:
                timestamp_pair.last_modified_dataset_props = (
                    dataset_properties_aspect.lastModified
                )

            if (
                dataset_operation_aspect
                and dataset_operation_aspect.lastUpdatedTimestamp
            ):
                timestamp_pair.last_updated_timestamp_dataset_props = max(
                    timestamp_pair.last_updated_timestamp_dataset_props or 0,
                    dataset_operation_aspect.lastUpdatedTimestamp,
                )

        else:
            # Create new TimestampPair
            last_modified_dataset_props: Optional[TimeStampClass] = None
            last_updated_timestamp_dataset_props: Optional[int] = None

            if dataset_properties_aspect:
                last_modified_dataset_props = dataset_properties_aspect.lastModified

            if dataset_operation_aspect:
                last_updated_timestamp_dataset_props = (
                    dataset_operation_aspect.lastUpdatedTimestamp
                )

            candidate_dataset_for_patch[wu.get_urn()] = TimestampPair(
                last_modified_dataset_props=last_modified_dataset_props,
                last_updated_timestamp_dataset_props=last_updated_timestamp_dataset_props,
            )

        yield wu

    # Emit a patch datasetProperties aspect for dataset where last_modified is None
    for entity_urn, timestamp_pair in candidate_dataset_for_patch.items():
        # Emit patch if last_modified is not set and last_updated_timestamp is set
        if (
            timestamp_pair.last_modified_dataset_props is None
            and timestamp_pair.last_updated_timestamp_dataset_props
        ):
            dataset_patch_builder = DatasetPatchBuilder(urn=entity_urn)

            dataset_patch_builder.set_last_modified(
                timestamp=TimeStampClass(
                    time=timestamp_pair.last_updated_timestamp_dataset_props
                )
            )

            yield from [
                MetadataWorkUnit(
                    id=MetadataWorkUnit.generate_workunit_id(mcp),
                    mcp_raw=mcp,
                )
                for mcp in dataset_patch_builder.build()
            ]



from typing import Union

from datahub.api.entities.assertion.field_assertion import (
    FieldMetricAssertion,
    FieldValuesAssertion,
)
from datahub.api.entities.assertion.freshness_assertion import (
    CronFreshnessAssertion,
    FixedIntervalFreshnessAssertion,
)
from datahub.api.entities.assertion.sql_assertion import (
    SqlMetricAssertion,
    SqlMetricChangeAssertion,
)
from datahub.api.entities.assertion.volume_assertion import (
    RowCountChangeVolumeAssertion,
    RowCountTotalVolumeAssertion,
)

# Pydantic v2 smart union: automatically discriminates based on the 'type' field
# (eg freshness/volume/sql/field) and unique fields within each type
DataHubAssertion = Union[
    FixedIntervalFreshnessAssertion,
    CronFreshnessAssertion,
    RowCountTotalVolumeAssertion,
    RowCountChangeVolumeAssertion,
    SqlMetricAssertion,
    SqlMetricChangeAssertion,
    FieldMetricAssertion,
    FieldValuesAssertion,
]



"""
Hive Metadata Processor - Shared WorkUnit generation for HMS connectors.

This module provides the HiveMetadataProcessor class that generates DataHub
MetadataWorkUnits from Hive Metastore data. It is data-fetcher agnostic and
works with any implementation of the HiveDataFetcher Protocol.

The processor handles:
- Table and view WorkUnit generation
- Container (catalog/database/schema) generation
- Schema field extraction with complex type support
- Storage lineage emission
- View lineage integration via SqlParsingAggregator
"""

import base64
import dataclasses
import json
import logging
from collections import namedtuple
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import (
    DatasetContainerSubTypes,
    DatasetSubTypes,
)
from datahub.ingestion.source.sql.hive.exceptions import InvalidDatasetIdentifierError
from datahub.ingestion.source.sql.hive.storage_lineage import HiveStorageLineage
from datahub.ingestion.source.sql.sql_common import SqlWorkUnit, get_schema_metadata
from datahub.ingestion.source.sql.sql_report import SQLSourceReport
from datahub.ingestion.source.sql.sql_utils import (
    add_table_to_schema_container,
    gen_database_container,
    gen_database_key,
    gen_schema_container,
    gen_schema_key,
    get_domain_wu,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import StatusClass
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import DatasetSnapshot
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import SchemaField
from datahub.metadata.schema_classes import (
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
    SubTypesClass,
    ViewPropertiesClass,
)
from datahub.utilities.groupby import groupby_unsorted
from datahub.utilities.hive_schema_to_avro import get_schema_fields_for_hive_column

if TYPE_CHECKING:
    from datahub.ingestion.source.sql.hive.hive_data_fetcher import HiveDataFetcher
    from datahub.ingestion.source.sql.hive.hive_metastore_source import (
        HiveMetastore,
        HiveMetastoreSourceReport,
    )
    from datahub.sql_parsing.sql_parsing_aggregator import SqlParsingAggregator
    from datahub.utilities.registries.domain_registry import DomainRegistry

logger = logging.getLogger(__name__)

TableKey = namedtuple("TableKey", ["schema", "table"])


@dataclasses.dataclass
class ViewDataset:
    """Container for view metadata during processing."""

    dataset_name: str
    schema_name: str
    columns: List[Dict[str, Any]]
    view_definition: Optional[str] = None


class HiveMetadataProcessor:
    """
    Processes Hive Metastore data and generates DataHub WorkUnits.

    This class is data-fetcher agnostic and works with any implementation
    of the HiveDataFetcher Protocol (SQLAlchemyDataFetcher, ThriftDataFetcher).

    It extracts shared metadata processing logic from HiveMetastoreSource
    to enable composition-based design.
    """

    # Presto view markers for base64-encoded view definitions
    _PRESTO_VIEW_PREFIX = "/* Presto View: "
    _PRESTO_VIEW_SUFFIX = " */"

    def __init__(
        self,
        config: "HiveMetastore",
        fetcher: "HiveDataFetcher",
        report: "HiveMetastoreSourceReport",
        platform: str,
        aggregator: Optional["SqlParsingAggregator"] = None,
        domain_registry: Optional["DomainRegistry"] = None,
    ):
        """
        Initialize the metadata processor.

        Args:
            config: HiveMetastore configuration
            fetcher: Data fetcher implementing HiveDataFetcher Protocol
            report: Source report for tracking progress and warnings
            platform: Platform name (e.g., "hive", "presto")
            aggregator: Optional SqlParsingAggregator for view lineage
            domain_registry: Optional domain registry for domain assignment
        """
        self.config = config
        self.fetcher = fetcher
        self.report = report
        self.platform = platform
        self.aggregator = aggregator
        self.domain_registry = domain_registry

        # Initialize subtypes based on config
        self.database_container_subtype = (
            DatasetContainerSubTypes.CATALOG
            if config.use_catalog_subtype
            else DatasetContainerSubTypes.DATABASE
        )
        self.view_subtype = (
            DatasetSubTypes.VIEW.title()
            if config.use_dataset_pascalcase_subtype
            else DatasetSubTypes.VIEW.lower()
        )
        self.table_subtype = (
            DatasetSubTypes.TABLE.title()
            if config.use_dataset_pascalcase_subtype
            else DatasetSubTypes.TABLE.lower()
        )

        # Initialize storage lineage handler
        self.storage_lineage = HiveStorageLineage(
            config=config,
            env=config.env,
        )

    def _get_db_name(self) -> str:
        """Get the database/catalog name for URN generation."""
        if self.config.catalog_name:
            return self.config.catalog_name
        if self.config.metastore_db_name:
            return self.config.metastore_db_name
        if self.config.database:
            return self.config.database
        return "hive"

    def _get_identifier(self, schema: str, entity: str) -> str:
        """Generate dataset identifier from schema and entity names."""
        return f"{schema}.{entity}"

    def _get_dataplatform_instance_aspect(
        self, dataset_urn: str
    ) -> Optional[MetadataWorkUnit]:
        """Generate DataPlatformInstance aspect if platform_instance is configured."""
        if self.config.platform_instance:
            return MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=DataPlatformInstanceClass(
                    platform=make_data_platform_urn(self.platform),
                    instance=make_dataplatform_instance_urn(
                        self.platform, self.config.platform_instance
                    ),
                ),
            ).as_workunit()
        return None

    def get_db_schema(self, dataset_identifier: str) -> Tuple[Optional[str], str]:
        """
        Extract database and schema from a dataset identifier.

        Args:
            dataset_identifier: Fully qualified dataset name (e.g., "catalog.db.table")

        Returns:
            Tuple of (database, schema) for view lineage context

        Raises:
            InvalidDatasetIdentifierError: If identifier is invalid
        """
        if dataset_identifier is None:
            raise InvalidDatasetIdentifierError("dataset_identifier cannot be None")
        elif not dataset_identifier.strip():
            raise InvalidDatasetIdentifierError("dataset_identifier cannot be empty")

        parts = dataset_identifier.split(".")
        parts = [p for p in parts if p]
        if not parts:
            raise InvalidDatasetIdentifierError(
                f"Invalid dataset identifier: {dataset_identifier}"
            )

        if self.config.include_catalog_name_in_ids:
            if len(parts) >= 3:
                return parts[0], parts[1]
            elif len(parts) == 2:
                return None, parts[0]
            else:
                return None, parts[0]
        else:
            if len(parts) >= 2:
                return None, parts[0]
            else:
                return None, parts[0]

    @staticmethod
    def _get_table_key(row: Dict[str, Any]) -> TableKey:
        """Extract table key (schema, table) from a row."""
        return TableKey(schema=row["schema_name"], table=row["table_name"])

    def get_schema_fields_for_column(
        self,
        column: Dict[str, Any],
    ) -> List[SchemaField]:
        """
        Generate SchemaField(s) for a column.

        Args:
            column: Column metadata dict with col_name, col_type, col_description

        Returns:
            List of SchemaField objects (may be multiple for complex types)
        """
        return get_schema_fields_for_hive_column(
            column["col_name"],
            column["col_type"],
            description=column.get("col_description", ""),
            default_nullable=True,
        )

    def get_schema_fields(
        self, dataset_name: str, columns: List[Dict[str, Any]]
    ) -> List[SchemaField]:
        """
        Generate all SchemaFields for a table/view.

        Args:
            dataset_name: Dataset name for logging
            columns: List of column metadata dicts

        Returns:
            List of all SchemaField objects
        """
        schema_fields: List[SchemaField] = []
        for column in columns:
            fields = self.get_schema_fields_for_column(column)
            schema_fields.extend(fields)
        return schema_fields

    @staticmethod
    def _set_partition_key(
        columns: List[Dict[str, Any]], schema_fields: List[SchemaField]
    ) -> None:
        """Mark partition columns in schema fields."""
        if not columns:
            return

        partition_key_names = {
            col["col_name"] for col in columns if col.get("is_partition_col")
        }

        for schema_field in schema_fields:
            name = schema_field.fieldPath.split(".")[-1]
            if name in partition_key_names:
                schema_field.isPartitioningKey = True

    def _get_presto_view_column_metadata(
        self, view_original_text: str
    ) -> Tuple[List[Dict[str, Any]], str]:
        """Extract column metadata from base64-encoded Presto view definition."""
        encoded_view_info = view_original_text.split(self._PRESTO_VIEW_PREFIX, 1)[
            -1
        ].rsplit(self._PRESTO_VIEW_SUFFIX, 1)[0]

        decoded_view_info = base64.b64decode(encoded_view_info)
        view_definition = json.loads(decoded_view_info).get("originalSql")

        columns = json.loads(decoded_view_info).get("columns")
        for col in columns:
            col["col_name"], col["col_type"] = col["name"], col["type"]

        return list(columns), view_definition

    # =========================================================================
    # WorkUnit Generation Methods
    # =========================================================================

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        """
        Generate all WorkUnits for tables, views, and containers.

        This is the main entry point for WorkUnit generation.
        """
        db_name = self._get_db_name()

        # Generate database container
        yield from self.gen_database_containers(db_name)

        # Generate schema containers
        yield from self.gen_schema_containers(db_name)

        # Process tables (respecting include_tables config)
        if self.config.include_tables:
            yield from self.process_tables(db_name)

        # Process views (respecting include_views config)
        if self.config.include_views:
            yield from self.process_views(db_name)

    def gen_database_containers(
        self,
        database: str,
        extra_properties: Optional[Dict[str, Any]] = None,
    ) -> Iterable[MetadataWorkUnit]:
        """Generate container WorkUnits for the database/catalog level."""
        database_container_key = gen_database_key(
            database,
            platform=self.platform,
            platform_instance=self.config.platform_instance,
            env=self.config.env,
        )

        yield from gen_database_container(
            database=database,
            database_container_key=database_container_key,
            sub_types=[self.database_container_subtype],
            domain_registry=self.domain_registry,
            domain_config=self.config.domain,
            extra_properties=extra_properties,
        )

    def gen_schema_containers(
        self,
        database: str,
        extra_properties: Optional[Dict[str, Any]] = None,
    ) -> Iterable[MetadataWorkUnit]:
        """Generate container WorkUnits for each schema/database."""
        for row in self.fetcher.fetch_schema_rows():
            schema = row["schema"]
            if not self.config.database_pattern.allowed(schema):
                continue

            database_container_key = gen_database_key(
                database,
                platform=self.platform,
                platform_instance=self.config.platform_instance,
                env=self.config.env,
            )

            schema_container_key = gen_schema_key(
                db_name=database,
                schema=schema,
                platform=self.platform,
                platform_instance=self.config.platform_instance,
                env=self.config.env,
            )

            yield from gen_schema_container(
                database=database,
                schema=schema,
                sub_types=[DatasetContainerSubTypes.SCHEMA],
                database_container_key=database_container_key,
                schema_container_key=schema_container_key,
                domain_registry=self.domain_registry,
                domain_config=self.config.domain,
                extra_properties=extra_properties,
            )

    def _add_dataset_to_container(
        self, dataset_urn: str, db_name: str, schema: str
    ) -> Iterable[MetadataWorkUnit]:
        """Add a dataset to its parent schema container."""
        schema_container_key = gen_schema_key(
            db_name=db_name,
            schema=schema,
            platform=self.platform,
            platform_instance=self.config.platform_instance,
            env=self.config.env,
        )
        yield from add_table_to_schema_container(
            dataset_urn=dataset_urn,
            parent_container_key=schema_container_key,
        )

    def _get_table_properties(self, db_name: str) -> Dict[str, Dict[str, str]]:
        """Build a cache of table properties keyed by dataset name."""
        table_properties: Dict[str, Dict[str, str]] = {}
        for row in self.fetcher.fetch_table_properties_rows():
            dataset_name = f"{row['schema_name']}.{row['table_name']}"
            if self.config.include_catalog_name_in_ids:
                dataset_name = f"{db_name}.{dataset_name}"
            if row["PARAM_KEY"] and row["PARAM_VALUE"]:
                table_properties.setdefault(dataset_name, {})[row["PARAM_KEY"]] = row[
                    "PARAM_VALUE"
                ]
        return table_properties

    def process_tables(
        self, db_name: str
    ) -> Iterable[Union[SqlWorkUnit, MetadataWorkUnit]]:
        """
        Process all tables and generate WorkUnits.

        Args:
            db_name: Database/catalog name for URN generation
        """
        # Build properties cache
        properties_cache = self._get_table_properties(db_name)

        # Fetch and process table rows
        table_rows = self.fetcher.fetch_table_rows()

        for key, group in groupby_unsorted(table_rows, self._get_table_key):
            schema_name = (
                f"{db_name}.{key.schema}"
                if self.config.include_catalog_name_in_ids
                else key.schema
            )

            dataset_name = self._get_identifier(schema=schema_name, entity=key.table)

            self.report.report_entity_scanned(dataset_name, ent_type="table")

            # Apply database pattern filter
            if not self.config.database_pattern.allowed(key.schema):
                self.report.report_dropped(f"{dataset_name}")
                continue

            # Apply table pattern filter
            if not self.config.table_pattern.allowed(dataset_name):
                self.report.report_dropped(dataset_name)
                continue

            columns = list(group)
            if len(columns) == 0:
                self.report.report_warning(dataset_name, "missing column information")

            dataset_urn: str = make_dataset_urn_with_platform_instance(
                self.platform,
                dataset_name,
                self.config.platform_instance,
                self.config.env,
            )
            dataset_snapshot = DatasetSnapshot(
                urn=dataset_urn,
                aspects=[StatusClass(removed=False)],
            )

            # Generate schema fields
            schema_fields = self.get_schema_fields(dataset_name, columns)
            self._set_partition_key(columns, schema_fields)

            schema_metadata = get_schema_metadata(
                cast(SQLSourceReport, self.report),
                dataset_name,
                self.platform,
                columns,
                None,
                None,
                schema_fields,
                self.config.simplify_nested_field_paths,
            )
            dataset_snapshot.aspects.append(schema_metadata)

            # Build properties
            properties: Dict[str, str] = properties_cache.get(dataset_name, {})
            properties["table_type"] = str(columns[-1]["table_type"] or "")
            properties["table_location"] = str(columns[-1]["table_location"] or "")
            properties["create_date"] = str(columns[-1]["create_date"] or "")

            par_columns: str = ", ".join(
                [c["col_name"] for c in columns if c["is_partition_col"]]
            )
            if par_columns != "":
                properties["partitioned_columns"] = par_columns

            table_description = properties.get("comment")

            # Add to container
            yield from self._add_dataset_to_container(
                dataset_urn=dataset_urn, db_name=db_name, schema=key.schema
            )

            # Emit properties
            if self.config.enable_properties_merge:
                from datahub.specific.dataset import DatasetPatchBuilder

                patch_builder: DatasetPatchBuilder = DatasetPatchBuilder(
                    urn=dataset_snapshot.urn
                )
                patch_builder.set_display_name(key.table)

                if table_description:
                    patch_builder.set_description(description=table_description)

                for prop, value in properties.items():
                    patch_builder.add_custom_property(key=prop, value=value)
                yield from [
                    MetadataWorkUnit(
                        id=f"{mcp_raw.entityUrn}-{DatasetPropertiesClass.ASPECT_NAME}",
                        mcp_raw=mcp_raw,
                    )
                    for mcp_raw in patch_builder.build()
                ]
            else:
                dataset_properties = DatasetPropertiesClass(
                    name=key.table,
                    description=table_description,
                    customProperties=properties,
                )
                dataset_snapshot.aspects.append(dataset_properties)

            mce = MetadataChangeEvent(proposedSnapshot=dataset_snapshot)
            yield SqlWorkUnit(id=dataset_name, mce=mce)

            # Emit subtype
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=SubTypesClass(typeNames=[self.table_subtype]),
            ).as_workunit()

            # Emit platform instance
            dpi_aspect = self._get_dataplatform_instance_aspect(dataset_urn)
            if dpi_aspect:
                yield dpi_aspect

            # Emit domain
            if self.config.domain and self.domain_registry:
                yield from get_domain_wu(
                    dataset_name=dataset_name,
                    entity_urn=dataset_urn,
                    domain_config=self.config.domain,
                    domain_registry=self.domain_registry,
                )

            # Emit storage lineage
            if self.config.emit_storage_lineage and properties.get("table_location"):
                table_dict = {
                    "StorageDescriptor": {"Location": properties["table_location"]}
                }
                yield from self.storage_lineage.get_lineage_mcp(
                    dataset_urn=dataset_urn,
                    table=table_dict,
                    dataset_schema=schema_metadata,
                )

    def _get_hive_view_datasets(self, db_name: str) -> Iterable[ViewDataset]:
        """Iterate over Hive view datasets from view rows."""
        view_rows = self.fetcher.fetch_view_rows()

        for key, group in groupby_unsorted(view_rows, self._get_table_key):
            schema_name = (
                f"{db_name}.{key.schema}"
                if self.config.include_catalog_name_in_ids
                else key.schema
            )

            dataset_name = self._get_identifier(schema=schema_name, entity=key.table)

            if not self.config.database_pattern.allowed(key.schema):
                self.report.report_dropped(f"{dataset_name}")
                continue

            columns = list(group)

            if len(columns) == 0:
                self.report.report_warning(dataset_name, "missing column information")

            yield ViewDataset(
                dataset_name=dataset_name,
                schema_name=key.schema,
                columns=columns,
                view_definition=columns[-1].get("view_expanded_text")
                if columns
                else None,
            )

    def _get_presto_view_datasets(self, db_name: str) -> Iterable[ViewDataset]:
        """
        Iterate over Presto/Trino view datasets.

        Note: This requires SQL fetcher as it needs to query VIEW_ORIGINAL_TEXT
        which contains base64-encoded view metadata.
        """
        from datahub.ingestion.source.sql.hive.hive_sql_fetcher import (
            SQLAlchemyDataFetcher,
        )

        if not isinstance(self.fetcher, SQLAlchemyDataFetcher):
            logger.warning(
                "Presto/Trino view extraction requires SQL connection. "
                "Skipping Presto views."
            )
            return

        for row in self.fetcher.fetch_presto_view_rows():
            schema_name = (
                f"{db_name}.{row['schema']}"
                if self.config.include_catalog_name_in_ids
                else row["schema"]
            )
            dataset_name = self._get_identifier(schema=schema_name, entity=row["name"])

            columns, view_definition = self._get_presto_view_column_metadata(
                row["view_original_text"]
            )

            if len(columns) == 0:
                self.report.report_warning(dataset_name, "missing column information")

            yield ViewDataset(
                dataset_name=dataset_name,
                schema_name=row["schema"],
                columns=columns,
                view_definition=view_definition,
            )

    def process_views(
        self, db_name: str
    ) -> Iterable[Union[SqlWorkUnit, MetadataWorkUnit]]:
        """
        Process all views and generate WorkUnits.

        Args:
            db_name: Database/catalog name for URN generation
        """
        from datahub.ingestion.source.sql.hive.hive_metastore_config import (
            HiveMetastoreConfigMode,
        )

        # Select view iterator based on mode
        if self.config.mode in [HiveMetastoreConfigMode.hive]:
            view_datasets = self._get_hive_view_datasets(db_name)
        else:
            view_datasets = self._get_presto_view_datasets(db_name)

        for dataset in view_datasets:
            self.report.report_entity_scanned(dataset.dataset_name, ent_type="view")

            # Apply view pattern filter
            if not self.config.view_pattern.allowed(dataset.dataset_name):
                self.report.report_dropped(dataset.dataset_name)
                continue

            dataset_urn = make_dataset_urn_with_platform_instance(
                self.platform,
                dataset.dataset_name,
                self.config.platform_instance,
                self.config.env,
            )
            dataset_snapshot = DatasetSnapshot(
                urn=dataset_urn,
                aspects=[StatusClass(removed=False)],
            )

            # Generate schema fields
            schema_fields = self.get_schema_fields(
                dataset.dataset_name, dataset.columns
            )

            schema_metadata = get_schema_metadata(
                cast(SQLSourceReport, self.report),
                dataset.dataset_name,
                self.platform,
                dataset.columns,
                canonical_schema=schema_fields,
                simplify_nested_field_paths=self.config.simplify_nested_field_paths,
            )
            dataset_snapshot.aspects.append(schema_metadata)

            # Add properties
            properties: Dict[str, str] = {"is_view": "True"}
            dataset_properties = DatasetPropertiesClass(
                name=dataset.dataset_name.split(".")[-1],
                description=None,
                customProperties=properties,
            )
            dataset_snapshot.aspects.append(dataset_properties)

            # Add view properties
            view_properties = ViewPropertiesClass(
                materialized=False,
                viewLogic=dataset.view_definition if dataset.view_definition else "",
                viewLanguage="SQL",
            )
            dataset_snapshot.aspects.append(view_properties)

            # Add to container
            yield from self._add_dataset_to_container(
                dataset_urn=dataset_urn, db_name=db_name, schema=dataset.schema_name
            )

            mce = MetadataChangeEvent(proposedSnapshot=dataset_snapshot)
            yield SqlWorkUnit(id=dataset.dataset_name, mce=mce)

            # Emit subtype
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=SubTypesClass(typeNames=[self.view_subtype]),
            ).as_workunit()

            # Emit platform instance
            dpi_aspect = self._get_dataplatform_instance_aspect(dataset_urn)
            if dpi_aspect:
                yield dpi_aspect

            # Emit view properties again via MCP
            view_properties_aspect = ViewPropertiesClass(
                materialized=False,
                viewLanguage="SQL",
                viewLogic=dataset.view_definition if dataset.view_definition else "",
            )
            yield MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=view_properties_aspect,
            ).as_workunit()

            # Emit domain
            if self.config.domain and self.domain_registry:
                yield from get_domain_wu(
                    dataset_name=dataset.dataset_name,
                    entity_urn=dataset_urn,
                    domain_registry=self.domain_registry,
                    domain_config=self.config.domain,
                )

            # Add view lineage
            if dataset.view_definition and self.config.include_view_lineage:
                if self.aggregator is not None:
                    default_db = None
                    default_schema = None
                    try:
                        default_db, default_schema = self.get_db_schema(
                            dataset.dataset_name
                        )
                    except InvalidDatasetIdentifierError as e:
                        logger.warning(
                            f"Invalid view identifier '{dataset.dataset_name}': {e}"
                        )
                        continue

                    self.aggregator.add_view_definition(
                        view_urn=dataset_urn,
                        view_definition=dataset.view_definition,
                        default_db=default_db,
                        default_schema=default_schema,
                    )



import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import datahub.emitter.mce_builder as builder
from datahub.emitter.aspect import ASPECT_MAP
from datahub.emitter.mce_builder import Aspect
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import ControlRecord, EndOfStream, RecordEnvelope
from datahub.ingestion.api.transform import Transformer
from datahub.metadata.schema_classes import (
    MetadataChangeEventClass,
    MetadataChangeProposalClass,
)
from datahub.utilities.urns.urn import Urn, guess_entity_type

log = logging.getLogger(__name__)


def _update_work_unit_id(
    envelope: RecordEnvelope, urn: str, aspect_name: str
) -> Dict[Any, Any]:
    structured_urn = Urn.from_string(urn)
    simple_name = "-".join(structured_urn.entity_ids)
    record_metadata = envelope.metadata.copy()
    record_metadata.update({"workunit_id": f"txform-{simple_name}-{aspect_name}"})
    return record_metadata


class HandleEndOfStreamTransformer:
    def handle_end_of_stream(
        self,
    ) -> Sequence[Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]]:
        return []


class LegacyMCETransformer(
    Transformer, HandleEndOfStreamTransformer, metaclass=ABCMeta
):
    @abstractmethod
    def transform_one(self, mce: MetadataChangeEventClass) -> MetadataChangeEventClass:
        pass


class SingleAspectTransformer(HandleEndOfStreamTransformer, metaclass=ABCMeta):
    @abstractmethod
    def aspect_name(self) -> str:
        """Implement this method to specify a single aspect that the transformer is interested in subscribing to. No default provided."""
        pass

    @abstractmethod
    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        """Implement this method to transform a single aspect for an entity.
        param: entity_urn: the entity that is being processed
        param: aspect_name: the aspect name corresponding to the subscription
        param: aspect: an optional aspect corresponding to the aspect name that the transformer is interested in. Empty if no aspect with this name was produced by the underlying connector
        """
        pass


class MultipleAspectTransformer(HandleEndOfStreamTransformer, metaclass=ABCMeta):
    """
    Transformer that transforms one input aspect into multiple output aspects.

    Unlike SingleAspectTransformer where input and output aspect types must match,
    this allows emitting additional aspects of different types.

    Example: Transform globalTags into both globalTags and structuredProperties.
    """

    @abstractmethod
    def aspect_name(self) -> str:
        """
        Specify the INPUT aspect this transformer subscribes to.
        Example: "globalTags"
        """
        pass

    @abstractmethod
    def transform_aspects(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Iterable[Tuple[str, Optional[Aspect]]]:
        """
        Transform one input aspect into multiple output aspects.

        Args:
            entity_urn: The entity being processed
            aspect_name: The input aspect name (from aspect_name() method)
            aspect: The input aspect value (None if not present)

        Returns:
            Iterable of (output_aspect_name, output_aspect) tuples.
            - First tuple conventionally contains the transformed input aspect
            - Additional tuples contain new aspects to emit
            - Use None as aspect value to remove that aspect

        Example:
            return [
                ("globalTags", modified_tags),      # Same type as input
                ("structuredProperties", props),    # Different type
            ]
        """
        pass


class BaseTransformer(Transformer, metaclass=ABCMeta):
    """
    Base transformer providing common functionality for all transformers.

    Architecture:
        BaseTransformer uses a mixin pattern where concrete transformers must inherit from
        both BaseTransformer and one of the allowed mixins:
        - LegacyMCETransformer: For full MCE transformation
        - SingleAspectTransformer: For transforming a single aspect type
        - MultipleAspectTransformer: For transforming one aspect into multiple aspects

        Example:
            class MyTransformer(BaseTransformer, SingleAspectTransformer):
                def aspect_name(self) -> str:
                    return "globalTags"
                def transform_aspect(self, ...) -> ...:
                    # transformation logic

        The mixins define the interface (abstract methods) while BaseTransformer provides
        the implementation helpers that dispatch based on which mixin is present. This allows
        type-safe polymorphic behavior with runtime assertions validating the class hierarchy.
    """

    allowed_mixins = [
        LegacyMCETransformer,
        SingleAspectTransformer,
        MultipleAspectTransformer,
    ]

    @abstractmethod
    def entity_types(self) -> List[str]:
        """Implement this method to specify which entity types the transformer is interested in subscribing to. Defaults to ALL (encoded as "*")"""
        return ["*"]

    def __init__(self):
        self.entity_map: Dict[str, Dict[str, Any]] = {}
        mixedin = False
        for mixin in [
            LegacyMCETransformer,
            SingleAspectTransformer,
            MultipleAspectTransformer,
        ]:
            mixedin = mixedin or isinstance(self, mixin)
        if not mixedin:
            assert (
                f"Class does not implement one of required traits {self.allowed_mixins}"
            )

    def _should_process(
        self,
        record: Union[
            MetadataChangeEventClass,
            MetadataChangeProposalWrapper,
            MetadataChangeProposalClass,
            ControlRecord,
        ],
    ) -> bool:
        if isinstance(record, ControlRecord):
            # all control events should be processed
            return True

        entity_types = self.entity_types()
        if "*" in entity_types:
            return True
        if isinstance(record, MetadataChangeEventClass):
            entity_type = guess_entity_type(record.proposedSnapshot.urn)
            return entity_type in entity_types
        elif isinstance(
            record, (MetadataChangeProposalWrapper, MetadataChangeProposalClass)
        ):
            return record.entityType in entity_types

        # default to process everything that is not caught by above checks
        return True

    def _record_mce(self, mce: MetadataChangeEventClass) -> None:
        record_entry = self.entity_map.get(mce.proposedSnapshot.urn, {"seen": {}})
        if "seen" in record_entry:
            # we just record the system metadata field from the mce, since we might need it later
            record_entry["seen"]["mce"] = mce.systemMetadata
            self.entity_map[mce.proposedSnapshot.urn] = record_entry

    def _record_mcp(
        self, mcp: Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]
    ) -> None:
        assert mcp.entityUrn
        record_entry = self.entity_map.get(mcp.entityUrn, {"seen": {}})
        if "seen" in record_entry and "mcp" not in record_entry["seen"]:
            # only record the first mcp seen
            record_entry["seen"]["mcp"] = mcp
        self.entity_map[mcp.entityUrn] = record_entry

    def _mark_processed(self, entity_urn: str) -> None:
        self.entity_map[entity_urn] = {"processed": True}

    def _transform_or_record_mce(
        self,
        envelope: RecordEnvelope[MetadataChangeEventClass],
    ) -> RecordEnvelope[MetadataChangeEventClass]:
        mce: MetadataChangeEventClass = envelope.record
        if mce.proposedSnapshot:
            self._record_mce(mce)
        if isinstance(self, SingleAspectTransformer):
            aspect_type = ASPECT_MAP[self.aspect_name()]

            # If we find a type corresponding to the aspect name we look for it in the mce
            # It's possible that the aspect is supported by the entity but not in the MCE
            # snapshot union. In those cases, we just want to record the urn as seen.
            supports_aspect = builder.can_add_aspect(mce, aspect_type)
            if supports_aspect:
                old_aspect = builder.get_aspect_if_available(
                    mce,
                    aspect_type,
                )
                if old_aspect is not None:
                    # TRICKY: If the aspect is not present in the MCE, it might still show up in a
                    # subsequent MCP. As such, we _only_ mark the urn as processed if we actually
                    # find the aspect already in the MCE.

                    transformed_aspect = self.transform_aspect(
                        entity_urn=mce.proposedSnapshot.urn,
                        aspect_name=self.aspect_name(),
                        aspect=old_aspect,
                    )

                    # If transformed_aspect is None, this will remove the aspect.
                    builder.set_aspect(
                        mce,
                        aspect_type=aspect_type,
                        aspect=transformed_aspect,
                    )

                    envelope.record = mce
                    self._mark_processed(mce.proposedSnapshot.urn)
        elif isinstance(self, MultipleAspectTransformer):
            input_aspect_type = ASPECT_MAP.get(self.aspect_name())
            if input_aspect_type and builder.can_add_aspect(mce, input_aspect_type):
                old_aspect = builder.get_aspect_if_available(mce, input_aspect_type)
                if old_aspect is not None:
                    for output_aspect_name, output_aspect in self.transform_aspects(
                        entity_urn=mce.proposedSnapshot.urn,
                        aspect_name=self.aspect_name(),
                        aspect=old_aspect,
                    ):
                        output_aspect_type = ASPECT_MAP.get(output_aspect_name)
                        if output_aspect_type and builder.can_add_aspect(
                            mce, output_aspect_type
                        ):
                            builder.set_aspect(
                                mce,
                                aspect_type=output_aspect_type,
                                aspect=output_aspect,
                            )
                    envelope.record = mce
                    self._mark_processed(mce.proposedSnapshot.urn)
        elif isinstance(self, LegacyMCETransformer):
            # we pass down the full MCE
            envelope.record = self.transform_one(mce)
            self._mark_processed(mce.proposedSnapshot.urn)

        return envelope

    def _transform_or_record_mcpw(
        self,
        envelope: RecordEnvelope[MetadataChangeProposalWrapper],
    ) -> Optional[RecordEnvelope[MetadataChangeProposalWrapper]]:
        """
        Handle SingleAspectTransformer processing for MetadataChangeProposalWrapper records.

        This helper method is defined on BaseTransformer but calls methods from the
        SingleAspectTransformer mixin. The runtime assertion ensures type safety while
        allowing the mixin pattern to work correctly.

        Args:
            envelope: Record envelope containing an MCPW

        Returns:
            The envelope with transformed aspect, or None if aspect should be dropped

        Note:
            This method requires self to be an instance of SingleAspectTransformer,
            which is validated via runtime assertion.
        """
        assert envelope.record.entityUrn
        assert isinstance(self, SingleAspectTransformer)
        if envelope.record.aspectName == self.aspect_name() and envelope.record.aspect:
            # we have a match on the aspect name, call the specific transform function
            transformed_aspect = self.transform_aspect(
                entity_urn=envelope.record.entityUrn,
                aspect_name=envelope.record.aspectName,
                aspect=envelope.record.aspect,
            )
            self._mark_processed(envelope.record.entityUrn)
            if transformed_aspect is None:
                # drop the record
                log.debug(
                    f"Dropping record {envelope} as transformation result is None"
                )
            envelope.record.aspect = transformed_aspect
        else:
            self._record_mcp(envelope.record)
        return envelope if envelope.record.aspect is not None else None

    def _transform_multiple_aspect_mcpw(
        self, envelope: RecordEnvelope[MetadataChangeProposalWrapper]
    ) -> Iterable[RecordEnvelope[MetadataChangeProposalWrapper]]:
        """
        Handle MultipleAspectTransformer processing for MetadataChangeProposalWrapper records.

        This helper method is defined on BaseTransformer but calls methods from the
        MultipleAspectTransformer mixin. It transforms a single input aspect into
        multiple output aspects, emitting them on-the-fly as separate envelopes.

        The first output aspect replaces the original envelope, and additional aspects
        are yielded as new envelopes with updated work unit IDs.

        Args:
            envelope: Record envelope containing an MCPW

        Yields:
            One or more envelopes, each containing a transformed aspect

        Note:
            This method requires self to be an instance of MultipleAspectTransformer,
            which is validated via runtime assertion.
        """
        assert isinstance(self, MultipleAspectTransformer)

        assert envelope.record.entityUrn
        if envelope.record.aspectName == self.aspect_name() and envelope.record.aspect:
            # Collect all output aspects
            output_aspects = list(
                self.transform_aspects(
                    entity_urn=envelope.record.entityUrn,
                    aspect_name=envelope.record.aspectName,
                    aspect=envelope.record.aspect,
                )
            )

            # First aspect replaces current envelope
            if output_aspects:
                first_aspect_name, first_aspect = output_aspects[0]
                if first_aspect is not None:
                    envelope.record.aspectName = first_aspect_name
                    envelope.record.aspect = first_aspect
                    # Update work unit ID for first aspect too
                    envelope.metadata = _update_work_unit_id(
                        envelope=envelope,
                        urn=envelope.record.entityUrn,
                        aspect_name=first_aspect_name,
                    )
                    yield envelope

                # Additional aspects yielded as new envelopes
                for aspect_name, aspect in output_aspects[1:]:
                    if aspect is not None:
                        record_metadata = _update_work_unit_id(
                            envelope=envelope,
                            urn=envelope.record.entityUrn,
                            aspect_name=aspect_name,
                        )
                        yield RecordEnvelope(
                            record=MetadataChangeProposalWrapper(
                                entityUrn=envelope.record.entityUrn,
                                aspectName=aspect_name,
                                aspect=aspect,
                            ),
                            metadata=record_metadata,
                        )
            self._mark_processed(envelope.record.entityUrn)
        else:
            self._record_mcp(envelope.record)

    def _handle_single_aspect_end_of_stream(
        self, envelope: RecordEnvelope
    ) -> Iterable[RecordEnvelope]:
        """
        Process accumulated state at end of stream for SingleAspectTransformer.

        This helper method is defined on BaseTransformer but calls methods from the
        SingleAspectTransformer mixin. It processes any entities that were tracked
        during the ingestion and emits their transformed aspects.

        This is necessary for transformers that need to see all aspects of an entity
        before deciding on the final transformation (e.g., accumulating state across
        multiple aspects).

        Args:
            envelope: The EndOfStream control record envelope

        Yields:
            Envelopes containing transformed aspects for tracked entities

        Note:
            This method requires self to be an instance of SingleAspectTransformer,
            which is validated via runtime assertion. It accesses entity_map which is
            managed by BaseTransformer.
        """
        assert isinstance(self, SingleAspectTransformer)

        for urn, state in self.entity_map.items():
            if "seen" in state:
                last_seen_mcp = state["seen"].get("mcp")
                last_seen_mce_system_metadata = state["seen"].get("mce")

                transformed_aspect = self.transform_aspect(
                    entity_urn=urn,
                    aspect_name=self.aspect_name(),
                    aspect=(
                        last_seen_mcp.aspect
                        if last_seen_mcp
                        and last_seen_mcp.aspectName == self.aspect_name()
                        else None
                    ),
                )
                if transformed_aspect:
                    structured_urn = Urn.from_string(urn)

                    mcp: MetadataChangeProposalWrapper = MetadataChangeProposalWrapper(
                        entityUrn=urn,
                        entityType=structured_urn.get_type(),
                        systemMetadata=(
                            last_seen_mcp.systemMetadata
                            if last_seen_mcp
                            else last_seen_mce_system_metadata
                        ),
                        aspectName=self.aspect_name(),
                        aspect=transformed_aspect,
                    )

                    if mcp.entityUrn:
                        assert mcp.aspect is not None
                        record_metadata = _update_work_unit_id(
                            envelope=envelope,
                            aspect_name=mcp.aspect.get_aspect_name(),
                            urn=mcp.entityUrn,
                        )
                    else:
                        record_metadata = envelope.metadata.copy()

                    yield RecordEnvelope(
                        record=mcp,
                        metadata=record_metadata,
                    )

            self._mark_processed(urn)

    def _handle_multiple_aspect_end_of_stream(
        self, envelope: RecordEnvelope
    ) -> Iterable[RecordEnvelope]:
        """
        Process accumulated state at end of stream for MultipleAspectTransformer.

        This helper method is defined on BaseTransformer but calls methods from the
        MultipleAspectTransformer mixin. It processes any entities that were tracked
        during the ingestion and emits their transformed aspects as multiple outputs.

        Similar to _handle_single_aspect_end_of_stream but supports emitting multiple
        aspects per entity (e.g., transforming tags into both tags and structured properties).

        Args:
            envelope: The EndOfStream control record envelope

        Yields:
            Envelopes containing transformed aspects for tracked entities. Multiple
            envelopes may be yielded per entity if the transformer emits multiple aspects.

        Note:
            This method requires self to be an instance of MultipleAspectTransformer,
            which is validated via runtime assertion. It accesses entity_map which is
            managed by BaseTransformer.
        """
        assert isinstance(self, MultipleAspectTransformer)

        for urn, state in self.entity_map.items():
            if "seen" in state:
                last_seen_mcp = state["seen"].get("mcp")
                last_seen_mce_system_metadata = state["seen"].get("mce")

                output_aspects = self.transform_aspects(
                    entity_urn=urn,
                    aspect_name=self.aspect_name(),
                    aspect=(
                        last_seen_mcp.aspect
                        if last_seen_mcp
                        and last_seen_mcp.aspectName == self.aspect_name()
                        else None
                    ),
                )

                for output_aspect_name, output_aspect in output_aspects:
                    if output_aspect:
                        structured_urn = Urn.from_string(urn)

                        mcp: MetadataChangeProposalWrapper = (
                            MetadataChangeProposalWrapper(
                                entityUrn=urn,
                                entityType=structured_urn.get_type(),
                                systemMetadata=(
                                    last_seen_mcp.systemMetadata
                                    if last_seen_mcp
                                    else last_seen_mce_system_metadata
                                ),
                                aspectName=output_aspect_name,
                                aspect=output_aspect,
                            )
                        )

                        if mcp.entityUrn:
                            record_metadata = _update_work_unit_id(
                                envelope=envelope,
                                aspect_name=output_aspect_name,
                                urn=mcp.entityUrn,
                            )
                        else:
                            record_metadata = envelope.metadata.copy()

                        yield RecordEnvelope(
                            record=mcp,
                            metadata=record_metadata,
                        )

            self._mark_processed(urn)

    def _handle_end_of_stream(
        self, envelope: RecordEnvelope
    ) -> Iterable[RecordEnvelope]:
        assert isinstance(
            self,
            (SingleAspectTransformer, MultipleAspectTransformer, LegacyMCETransformer),
        ), (
            f"Transformer {type(self).__name__} must inherit from one of the allowed mixins"
        )

        mcps: Sequence[
            Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]
        ] = self.handle_end_of_stream()

        for mcp in mcps:
            if (
                mcp.aspect is None or mcp.aspectName is None or mcp.entityUrn is None
            ):  # to silent the lint error
                continue

            record_metadata = _update_work_unit_id(
                envelope=envelope,
                aspect_name=mcp.aspectName,
                urn=mcp.entityUrn,
            )

            yield RecordEnvelope(
                record=mcp,
                metadata=record_metadata,
            )

    def transform(
        self, record_envelopes: Iterable[RecordEnvelope]
    ) -> Iterable[RecordEnvelope]:
        for envelope in record_envelopes:
            if not self._should_process(envelope.record):
                # early exit
                pass
            elif isinstance(envelope.record, MetadataChangeEventClass) and isinstance(
                self,
                (
                    SingleAspectTransformer,
                    MultipleAspectTransformer,
                    LegacyMCETransformer,
                ),
            ):
                envelope = self._transform_or_record_mce(envelope)
            elif isinstance(
                envelope.record, MetadataChangeProposalWrapper
            ) and isinstance(self, SingleAspectTransformer):
                return_envelope = self._transform_or_record_mcpw(envelope)
                if return_envelope is None:
                    continue
                else:
                    envelope = return_envelope
            elif isinstance(
                envelope.record, MetadataChangeProposalWrapper
            ) and isinstance(self, MultipleAspectTransformer):
                # Use helper method for MultipleAspectTransformer MCPW processing
                yield from self._transform_multiple_aspect_mcpw(envelope)
                continue
            elif isinstance(envelope.record, EndOfStream) and isinstance(
                self, SingleAspectTransformer
            ):
                # Use helper method for SingleAspectTransformer end of stream
                yield from self._handle_single_aspect_end_of_stream(envelope)
                yield from self._handle_end_of_stream(envelope=envelope)
            elif isinstance(envelope.record, EndOfStream) and isinstance(
                self, MultipleAspectTransformer
            ):
                # Use helper method for MultipleAspectTransformer end of stream
                yield from self._handle_multiple_aspect_end_of_stream(envelope)
                yield from self._handle_end_of_stream(envelope=envelope)

            yield envelope



"""SQLAlchemy event listeners for recording and replaying database queries.

This module provides event listeners that intercept SQLAlchemy query execution
at the engine level, avoiding the import reference issues with wrapper-based approaches.

Event listeners are attached to SQLAlchemy engines after creation, making them
immune to modules that import create_engine directly.
"""

import logging
from typing import Any, Dict, Optional, Tuple

from sqlalchemy.engine import Engine

from datahub.ingestion.recording.db_proxy import QueryRecorder, QueryRecording

logger = logging.getLogger(__name__)

# Thread-local storage for active recorders per engine
# Key: engine id, Value: (recorder, is_replay)
_active_recorders: Dict[int, Tuple[QueryRecorder, bool]] = {}


def attach_sqlalchemy_recorder(
    engine: Engine, recorder: QueryRecorder, is_replay: bool = False
) -> None:
    """Attach event listeners to a SQLAlchemy engine for recording/replay.

    Args:
        engine: SQLAlchemy engine to attach listeners to
        recorder: QueryRecorder instance for storing/retrieving recordings
        is_replay: If True, serve from recordings. If False, record queries.
    """
    engine_id = id(engine)
    _active_recorders[engine_id] = (recorder, is_replay)

    if is_replay:
        # Replay mode: use raw_connection wrapper (handled in patcher.py)
        # Event listeners are not used for replay to avoid execution
        logger.debug(
            "Replay mode: raw_connection wrapper will be used instead of event listeners"
        )
    else:
        # Recording mode: wrap engine.connect() to wrap connections when they're created
        original_connect = engine.connect

        def wrapped_connect(*args: Any, **kwargs: Any) -> Any:
            conn = original_connect(*args, **kwargs)
            return _wrap_connection_for_recording(conn)

        # Dynamically patch engine.connect() - intentional method assignment
        engine.connect = wrapped_connect  # type: ignore[method-assign]
        logger.debug(f"Wrapped engine.connect() for recording on engine {engine_id}")


def detach_sqlalchemy_recorder(engine: Engine) -> None:
    """Remove event listeners from a SQLAlchemy engine.

    Args:
        engine: SQLAlchemy engine to detach listeners from
    """
    engine_id = id(engine)
    if engine_id in _active_recorders:
        # Remove listeners (SQLAlchemy doesn't provide direct removal, but we can clear our reference)
        _active_recorders.pop(engine_id, None)
        logger.debug(f"Detached SQLAlchemy listeners from engine {engine_id}")


def _get_recorder_for_engine(conn: Any) -> Optional[Tuple[QueryRecorder, bool]]:
    """Get the active recorder for a connection's engine.

    Args:
        conn: SQLAlchemy connection object

    Returns:
        Tuple of (recorder, is_replay) or None if not found
    """
    # Get engine from connection
    engine = getattr(conn, "engine", None)
    if engine is None:
        return None

    engine_id = id(engine)
    return _active_recorders.get(engine_id)


def _get_query_string(statement: Any) -> str:
    """Extract query string from SQLAlchemy statement or string.

    Args:
        statement: SQLAlchemy statement object or string

    Returns:
        Query string
    """
    if hasattr(statement, "compile"):
        # SQLAlchemy statement object
        try:
            compiled = statement.compile(compile_kwargs={"literal_binds": True})
            return str(compiled)
        except Exception:
            # Fallback if compilation fails
            return str(statement)
    elif isinstance(statement, str):
        return statement
    else:
        return str(statement)


def _materialize_results(result: Any) -> list:
    """Materialize SQLAlchemy Result object into list of dictionaries.

    Args:
        result: SQLAlchemy Result object

    Returns:
        List of dictionaries representing rows
    """
    # Materialize results (this consumes the Result object)
    rows = list(result)

    if not rows:
        return []

    # Try to get column names from result keys (SQLAlchemy 2.x Row objects)
    if hasattr(rows[0], "_mapping"):
        # SQLAlchemy 2.x Row._mapping
        return [dict(row._mapping) for row in rows]
    elif hasattr(rows[0], "keys"):
        # SQLAlchemy 1.x RowProxy
        return [dict(row) for row in rows]

    # Tuple rows - need description
    description = getattr(result, "cursor_description", None)
    if description:
        column_names = [desc[0] for desc in description]
        return [dict(zip(column_names, row)) for row in rows]

    # No description - use indices
    return [dict(enumerate(row)) for row in rows]


class _ResultIterator:
    """Iterator that wraps materialized SQLAlchemy results."""

    def __init__(self, rows_data: list, original_result: Any):
        self._rows = rows_data
        self._index = 0
        self._original_result = original_result

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self._rows):
            row = self._rows[self._index]
            self._index += 1
            return row
        raise StopIteration

    def fetchall(self):
        """Fetch all rows from the result set.

        Returns:
            List of all rows.
        """
        return self._rows

    def all(self):
        """Return all rows from the result set (SQLAlchemy 2.x style).

        Returns:
            List of all rows.
        """
        return self._rows

    def first(self):
        """Return the first row or None if no rows.

        Returns:
            First row or None if result set is empty.
        """
        return self._rows[0] if self._rows else None

    def one(self):
        """Return exactly one row, raising ValueError if not exactly one.

        Returns:
            The single row.

        Raises:
            ValueError: If result set does not contain exactly one row.
        """
        if len(self._rows) == 1:
            return self._rows[0]
        raise ValueError("Expected exactly one row")

    def one_or_none(self):
        """Return one row or None if no rows, raising ValueError if multiple rows.

        Returns:
            The single row, or None if no rows.

        Raises:
            ValueError: If result set contains more than one row.
        """
        if len(self._rows) == 1:
            return self._rows[0]
        return None

    def scalar(self):
        """Return the first column of the first row, or None if no rows.

        Returns:
            Scalar value from first column of first row, or None if empty.
        """
        if self._rows and len(self._rows) > 0:
            first_row = self._rows[0]
            if hasattr(first_row, "__getitem__"):
                return first_row[0]
            return first_row
        return None

    # Preserve other attributes from original result
    def __getattr__(self, name):
        return getattr(self._original_result, name)


def _create_wrapped_execute(original_execute: Any, recorder: QueryRecorder) -> Any:
    """Create a wrapped execute function that records queries and results.

    Args:
        original_execute: Original connection.execute method
        recorder: QueryRecorder instance

    Returns:
        Wrapped execute function
    """

    def wrapped_execute(statement: Any, *args: Any, **kwargs: Any) -> Any:
        query = _get_query_string(statement)
        result = original_execute(statement, *args, **kwargs)

        try:
            rows = list(result)
            results = _materialize_results(result) if rows else []

            recording = QueryRecording(
                query=query,
                parameters=kwargs.get("parameters") or (args[0] if args else None),
                results=results,
                row_count=len(results),
            )
            recorder.record(recording)

            logger.debug(f"Recorded SQL query: {query[:100]}... ({len(results)} rows)")

            return _ResultIterator(rows, result)

        except Exception as e:
            logger.error(f"Error recording query: {e}", exc_info=True)
            recording = QueryRecording(
                query=query,
                parameters=kwargs.get("parameters") or (args[0] if args else None),
                error=str(e),
            )
            recorder.record(recording)
            return result

    return wrapped_execute


def _wrap_connection_for_recording(conn: Any) -> Any:
    """Wrap connection.execute() to capture results.

    This is called when a connection is created from engine.connect().
    We wrap the execute method to intercept queries and materialize results.
    """
    engine = getattr(conn, "engine", None)
    if not engine:
        logger.debug("No engine found on connection - skipping wrap")
        return conn

    engine_id = id(engine)
    recorder_info = _active_recorders.get(engine_id)
    if not recorder_info:
        logger.debug(f"No recorder found for engine {engine_id} - skipping wrap")
        return conn

    recorder, is_replay = recorder_info
    if is_replay:
        return conn

    if not hasattr(conn, "execute"):
        logger.warning("Connection does not have execute() method - cannot wrap")
        return conn

    original_execute = conn.execute
    logger.info("✅ Wrapping connection.execute() for recording")
    # Dynamically patch connection.execute() - intentional method assignment
    conn.execute = _create_wrapped_execute(original_execute, recorder)  # type: ignore[method-assign]

    return conn


# Note: We use connection.execute() wrapping instead of cursor event listeners
# because SQLAlchemy 2.x Result objects are not accessible in after_cursor_execute.
# Wrapping execute() allows us to materialize results from the Result object.


# Note: Replay mode uses raw_connection wrapper instead of event listeners
# to prevent actual database connections. Event listeners are only used for recording.



"""
Thrift client for Hive Metastore with Kerberos/SASL authentication support.

This module provides a wrapper around pymetastore's ThriftHiveMetastore client
with support for Kerberos authentication via SASL/GSSAPI, including hostname
override for environments where the connection hostname differs from the
Kerberos principal hostname (e.g., when connecting through a load balancer).

NOTE: Retry Logic Implementation
--------------------------------
This module uses `tenacity` for retry logic, which is consistent with other
DataHub connectors (slack, mode). Tenacity is included as an extra dependency.
Install with: pip install 'acryl-datahub[hive-metastore-thrift]'
"""

import logging
from dataclasses import dataclass
from datetime import datetime
from types import SimpleNamespace, TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
)

from pymetastore.hive_metastore import ThriftHiveMetastore
from pymetastore.hive_metastore.ttypes import GetTableRequest
from tenacity import (
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
)
from thrift.protocol import TBinaryProtocol
from thrift.Thrift import TException
from thrift.transport import TSocket, TTransport

if TYPE_CHECKING:
    from thrift_sasl import TSaslClientTransport

logger = logging.getLogger(__name__)

# =============================================================================
# Constants
# =============================================================================

# Hive table type for views
VIRTUAL_VIEW_TYPE = "VIRTUAL_VIEW"


# =============================================================================
# Dataclasses for failure tracking keys
# =============================================================================


@dataclass(frozen=True)
class DatabaseKey:
    """Key for tracking database-level failures."""

    catalog_name: Optional[str]
    db_name: str


@dataclass(frozen=True)
class TableKey:
    """Key for tracking table-level failures and caching."""

    catalog_name: Optional[str]
    db_name: str
    table_name: str


# Exceptions that indicate a catalog API is not supported (HMS 2.x)
# These should trigger fallback to non-catalog APIs.
#
# FALLBACK STRATEGY: When a catalog-aware API fails with one of these exceptions,
# we fall back to the non-catalog HMS 2.x API.
#
# IMPORTANT: We intentionally do NOT catch TTransportException here because:
# 1. Network errors should propagate so users know there's a connectivity issue
# 2. Silently falling back on network errors could mask real problems
# 3. Only "method not found" type errors should trigger fallback
_CATALOG_NOT_SUPPORTED_EXCEPTIONS = (
    AttributeError,  # Method doesn't exist on client
    TException,  # Thrift "method not found" or protocol errors (but not transport errors)
)

# More specific exception for when we need to distinguish API not found vs other errors
_CATALOG_API_NOT_FOUND_EXCEPTIONS = (
    AttributeError,  # Method doesn't exist on client
)


class HMSConnectionError(Exception):
    """Custom exception for HMS connection errors with helpful messages."""

    pass


def _wrap_thrift_error(e: Exception, use_kerberos: bool) -> Exception:
    """
    Wrap Thrift errors with more helpful messages.

    Args:
        e: The original exception
        use_kerberos: Whether Kerberos auth is enabled in config

    Returns:
        A more descriptive exception or the original exception
    """
    error_str = str(e).lower()

    # Check for auth-related connection errors
    is_auth_error = (
        isinstance(e, BrokenPipeError)
        or "broken pipe" in error_str
        or "read 0 bytes" in error_str  # EOF - server closed connection
        or "end of file" in error_str
        or "connection reset" in error_str
    )

    if is_auth_error:
        if not use_kerberos:
            return HMSConnectionError(
                f"Connection closed by HMS server. "
                f"This typically means the Hive Metastore requires SASL/Kerberos authentication, "
                f"but 'use_kerberos' is set to False. "
                f"Try setting 'use_kerberos: true' and configuring 'kerberos_service_name'. "
                f"Original error: {e}"
            )
        else:
            return HMSConnectionError(
                f"Connection closed by HMS server. "
                f"Kerberos is enabled but the connection failed. Check that: "
                f"1) You have a valid Kerberos ticket (run 'klist' to verify) "
                f"2) The 'kerberos_service_name' matches the HMS principal "
                f"3) Network connectivity to the HMS server is stable. "
                f"Original error: {e}"
            )

    return e


class ThriftInspectorAdapter:
    """
    Minimal adapter that provides an Inspector-like interface for HiveMetastoreSource.

    HiveMetastoreSource uses Inspector primarily to extract the database name
    via `inspector.engine.url.database`. This adapter provides just enough
    structure to satisfy that interface.
    """

    def __init__(self, database: str):
        """
        Create an adapter that mimics SQLAlchemy Inspector for database name access.

        Args:
            database: The database/catalog name to return
        """
        # Create a fake engine.url.database structure that HiveMetastoreSource expects
        self.engine = SimpleNamespace(url=SimpleNamespace(database=database))


@dataclass
class ThriftConnectionConfig:
    """Configuration for HMS Thrift client connection."""

    host: str
    port: int = 9083
    use_kerberos: bool = True
    kerberos_service_name: str = "hive"
    kerberos_hostname_override: Optional[str] = None
    timeout_seconds: int = 60
    max_retries: int = (
        3  # Uses exponential backoff internally (multiplier=2, max_wait=30s)
    )


class HiveMetastoreThriftClient:
    """
    Client for connecting to Hive Metastore via Thrift protocol.

    Supports both plain and Kerberized (SASL/GSSAPI) connections with
    automatic retry logic for transient failures.
    """

    def __init__(self, config: ThriftConnectionConfig):
        self.config = config
        self._transport: Optional[TTransport.TBufferedTransport] = None
        self._client: Optional[ThriftHiveMetastore.Client] = None
        # Track failures for reporting - use dicts to avoid duplicates
        # (both iter_table_rows and iter_table_properties_rows call _iter_tables_and_views)
        self._database_failures: Dict[DatabaseKey, str] = {}
        self._table_failures: Dict[TableKey, str] = {}
        # Cache table info to avoid duplicate API calls
        self._table_cache: Dict[TableKey, Dict[str, Any]] = {}

    def get_database_failures(self) -> List[tuple[str, str]]:
        """Get list of database failures as (db_name, error_message) tuples.

        Note: For multi-catalog scenarios, the db_name may be prefixed with catalog.
        """
        result: List[tuple[str, str]] = []
        for key, err in self._database_failures.items():
            # Include catalog in db_name for clarity if present
            display_name = (
                f"{key.catalog_name}.{key.db_name}" if key.catalog_name else key.db_name
            )
            result.append((display_name, err))
        return result

    def get_table_failures(self) -> List[tuple[str, str, str]]:
        """Get list of table failures as (db_name, table_name, error_message) tuples.

        Note: For multi-catalog scenarios, the db_name may be prefixed with catalog.
        """
        result: List[tuple[str, str, str]] = []
        for key, err in self._table_failures.items():
            # Include catalog in db_name for clarity if present
            display_name = (
                f"{key.catalog_name}.{key.db_name}" if key.catalog_name else key.db_name
            )
            result.append((display_name, key.table_name, err))
        return result

    def clear_failures(self) -> None:
        """Clear tracked failures and table cache."""
        self._database_failures.clear()
        self._table_failures.clear()
        self._table_cache.clear()

    def _create_sasl_transport(self, socket: TSocket.TSocket) -> "TSaslClientTransport":
        """Create SASL transport for Kerberos authentication."""
        try:
            from pyhive.sasl_compat import PureSASLClient
            from thrift_sasl import TSaslClientTransport
        except ImportError as e:
            raise ImportError(
                "Kerberos authentication requires 'thrift-sasl' and 'acryl-pyhive[hive-pure-sasl]'. "
                f"Install with: pip install thrift-sasl acryl-pyhive[hive-pure-sasl]. Error: {e}"
            ) from e

        # Use hostname override if provided, otherwise use connection host
        # This handles the case where the Kerberos principal hostname differs
        # from the actual connection hostname (e.g., load balancer scenarios)
        sasl_host = self.config.kerberos_hostname_override or self.config.host

        logger.debug(
            f"Setting up SASL/GSSAPI transport with service={self.config.kerberos_service_name}, "
            f"host={sasl_host} (connection host={self.config.host})"
        )

        def sasl_factory() -> PureSASLClient:
            return PureSASLClient(
                sasl_host,
                service=self.config.kerberos_service_name,
                mechanism="GSSAPI",
            )

        return TSaslClientTransport(sasl_factory, "GSSAPI", socket)

    def connect(self) -> None:
        """Establish connection to HMS."""
        logger.info(f"Connecting to HMS at {self.config.host}:{self.config.port}")

        socket = TSocket.TSocket(self.config.host, self.config.port)
        socket.setTimeout(self.config.timeout_seconds * 1000)

        if self.config.use_kerberos:
            self._transport = self._create_sasl_transport(socket)
        else:
            self._transport = TTransport.TBufferedTransport(socket)

        self._transport.open()

        protocol = TBinaryProtocol.TBinaryProtocol(self._transport)
        self._client = ThriftHiveMetastore.Client(protocol)

        logger.info("Successfully connected to HMS")

    def close(self) -> None:
        """Close the connection."""
        if self._transport is not None:
            try:
                self._transport.close()
            except Exception as e:
                logger.warning(f"Error closing transport: {e}")
            self._transport = None
            self._client = None

    def __enter__(self) -> "HiveMetastoreThriftClient":
        self.connect()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    @property
    def client(self) -> "ThriftHiveMetastore.Client":
        """Get the underlying Thrift client."""
        if self._client is None:
            raise RuntimeError("Client not connected. Call connect() first.")
        return self._client

    def _get_retry_decorator(
        self,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """Create a retry decorator with exponential backoff.

        Note: BrokenPipeError is NOT retried as it typically indicates
        an authentication mismatch (e.g., HMS requires SASL but client
        didn't use it). Such errors are wrapped with helpful messages.
        """

        def should_retry(exception: BaseException) -> bool:
            """Determine if an exception should trigger a retry."""
            # Don't retry on BrokenPipeError - it's not transient
            if isinstance(exception, BrokenPipeError):
                return False
            # Check for broken pipe in nested exceptions
            if hasattr(exception, "inner") and isinstance(
                exception.inner, BrokenPipeError
            ):
                return False
            # Retry on transport/connection errors
            return isinstance(
                exception,
                (TTransport.TTransportException, ConnectionError, TimeoutError),
            )

        return retry(
            retry=retry_if_exception(should_retry),
            stop=stop_after_attempt(self.config.max_retries),
            wait=wait_exponential(multiplier=2, max=30),  # Sensible defaults
            reraise=True,
        )

    def _call_with_error_handling(
        self, func: Callable[[], Any], operation_name: str = "HMS operation"
    ) -> Any:
        """
        Call a function with error handling and helpful error messages.

        Args:
            func: The function to call
            operation_name: Description of the operation for error messages

        Returns:
            The result of the function call

        Raises:
            HMSConnectionError: With helpful message for connection issues
            Other exceptions: Re-raised with original context
        """
        try:
            return func()
        except Exception as e:
            # Check if this looks like an auth-related connection error
            error_str = str(e).lower()
            is_auth_error = (
                isinstance(e, BrokenPipeError)
                or "broken pipe" in error_str
                or "read 0 bytes" in error_str
                or "end of file" in error_str
                or "connection reset" in error_str
            )

            if is_auth_error:
                raise _wrap_thrift_error(e, self.config.use_kerberos) from e

            # Check for BrokenPipeError in the exception chain
            current: Optional[BaseException] = e
            while current is not None:
                if isinstance(current, BrokenPipeError):
                    raise _wrap_thrift_error(current, self.config.use_kerberos) from e
                # Check TTransportException's inner attribute
                if hasattr(current, "inner"):
                    inner = getattr(current, "inner", None)
                    inner_str = str(inner).lower() if inner else ""
                    if (
                        isinstance(inner, BrokenPipeError)
                        or "read 0 bytes" in inner_str
                    ):
                        # inner is a known Exception type here
                        error_to_wrap: Exception = (
                            inner if isinstance(inner, Exception) else e
                        )
                        raise _wrap_thrift_error(
                            error_to_wrap,
                            self.config.use_kerberos,
                        ) from e
                    current = inner
                else:
                    current = current.__cause__
            # Re-raise original exception if not a wrapped error type
            raise

    # -------------------------------------------------------------------------
    # HMS API Methods - return data in dict format matching hive_metastore.py
    # With retry logic for transient failures
    # -------------------------------------------------------------------------

    def supports_catalogs(self) -> bool:
        """
        Check if the HMS server supports catalogs (HMS 3.x feature).

        Returns True if get_catalogs() returns a non-empty list,
        False otherwise (HMS 2.x or catalog feature disabled).

        Note: Only catches AttributeError (method doesn't exist on HMS 2.x).
        Network/auth errors will propagate up to caller.
        """
        try:
            catalogs = self.get_catalogs()
            return len(catalogs) > 0
        except AttributeError:
            # HMS 2.x doesn't have get_catalogs method
            logger.debug("HMS 2.x detected - catalogs not supported")
            return False

    def get_all_databases(self, catalog_name: Optional[str] = None) -> List[str]:
        """
        Get all database names with retry logic.

        Args:
            catalog_name: Optional catalog name for HMS 3.x. If provided and HMS
                          supports catalogs, uses catalog-aware pattern syntax.

        For HMS 3.x, databases in a specific catalog can be listed using the
        pattern syntax: @{catalog_name}# (e.g., @spark_catalog#)
        """

        @self._get_retry_decorator()
        def _call() -> List[str]:
            if catalog_name:
                try:
                    # HMS 3.x uses pattern @{catalog_name}# to list databases in a catalog
                    pattern = f"@{catalog_name}#"
                    result = self.client.get_databases(pattern)
                    return list(result) if result else []
                except _CATALOG_NOT_SUPPORTED_EXCEPTIONS as e:
                    logger.debug(
                        f"Catalog-aware get_databases failed for catalog '{catalog_name}', "
                        f"falling back to default: {e}"
                    )
                    return self.client.get_all_databases()
            return self.client.get_all_databases()

        return self._call_with_error_handling(_call, "get_all_databases")

    def get_database(
        self, db_name: str, catalog_name: Optional[str] = None
    ) -> Dict[str, Any]:
        """Get database metadata with retry logic."""

        @self._get_retry_decorator()
        def _call() -> Dict[str, Any]:
            if catalog_name:
                try:
                    # HMS 3.x API with catalog support
                    db = self.client.get_database_req(
                        {"name": db_name, "catalogName": catalog_name}
                    )
                except _CATALOG_NOT_SUPPORTED_EXCEPTIONS as e:
                    logger.debug(
                        f"Catalog-aware get_database failed, falling back: {e}"
                    )
                    db = self.client.get_database(db_name)
            else:
                db = self.client.get_database(db_name)
            return {
                "name": db.name,
                "description": db.description,
                "location": db.locationUri,
                "parameters": dict(db.parameters) if db.parameters else {},
            }

        return _call()

    def get_all_tables(
        self, db_name: str, catalog_name: Optional[str] = None
    ) -> List[str]:
        """
        Get all table names in a database with retry logic.

        Args:
            db_name: Database name
            catalog_name: Optional catalog name for HMS 3.x

        For HMS 3.x with catalog support, uses the pattern syntax
        @{catalog_name}#{db_name} to specify the catalog context.
        Falls back to standard get_all_tables() for HMS 2.x.
        """

        @self._get_retry_decorator()
        def _call() -> List[str]:
            if catalog_name:
                try:
                    # HMS 3.x uses pattern @{catalog}#{db} to list tables in a catalog
                    pattern = f"@{catalog_name}#{db_name}"
                    result = self.client.get_all_tables(pattern)
                    return list(result) if result else []
                except _CATALOG_NOT_SUPPORTED_EXCEPTIONS as e:
                    logger.debug(
                        f"Catalog-aware get_all_tables failed for catalog '{catalog_name}', "
                        f"falling back to default: {e}"
                    )
                    return self.client.get_all_tables(db_name)
            return self.client.get_all_tables(db_name)

        return _call()

    def get_table(
        self, db_name: str, table_name: str, catalog_name: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Get table metadata in a format compatible with hive_metastore.py.
        Includes retry logic for transient failures.

        Args:
            db_name: Database name
            table_name: Table name
            catalog_name: Optional catalog name for HMS 3.x
        """

        @self._get_retry_decorator()
        def _call() -> Dict[str, Any]:
            if catalog_name:
                try:
                    # HMS 3.x API - get_table_req with catalog
                    req = GetTableRequest(
                        catName=catalog_name, dbName=db_name, tblName=table_name
                    )
                    response = self.client.get_table_req(req)
                    table = response.table
                except _CATALOG_NOT_SUPPORTED_EXCEPTIONS as e:
                    logger.debug(
                        f"Catalog-aware get_table failed for catalog '{catalog_name}', "
                        f"falling back: {e}"
                    )
                    table = self.client.get_table(db_name, table_name)
            else:
                table = self.client.get_table(db_name, table_name)

            # Get table location from storage descriptor
            table_location = ""
            if table.sd and table.sd.location:
                table_location = table.sd.location

            # Extract table parameters
            parameters = dict(table.parameters) if table.parameters else {}

            return {
                "db_name": db_name,
                "table_name": table.tableName,
                "table_type": table.tableType,
                "create_time": table.createTime,
                "location": table_location,
                "parameters": parameters,
                "view_original_text": table.viewOriginalText,
                "view_expanded_text": table.viewExpandedText,
                "partition_keys": [
                    {
                        "col_name": pk.name,
                        "col_type": pk.type,
                        "col_description": pk.comment or "",
                    }
                    for pk in (table.partitionKeys or [])
                ],
            }

        return _call()

    def get_fields(
        self, db_name: str, table_name: str, catalog_name: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Get column/field metadata for a table with retry logic.

        Args:
            db_name: Database name
            table_name: Table name
            catalog_name: Optional catalog name for HMS 3.x

        Note: The Thrift get_fields() API doesn't have a catalog-aware variant.
        For HMS 3.x with catalog context, we extract columns from the table's
        StorageDescriptor (fetched via get_table_req with catalog).
        """

        @self._get_retry_decorator()
        def _call() -> List[Dict[str, Any]]:
            if catalog_name:
                # For HMS 3.x with catalog: get columns from table's StorageDescriptor
                # This ensures we get columns from the correct catalog context
                try:
                    req = GetTableRequest(
                        catName=catalog_name, dbName=db_name, tblName=table_name
                    )
                    response = self.client.get_table_req(req)
                    table = response.table

                    # Extract columns from StorageDescriptor
                    if table.sd and table.sd.cols:
                        return [
                            {
                                "col_name": col.name,
                                "col_type": col.type,
                                "col_description": col.comment or "",
                            }
                            for col in table.sd.cols
                        ]
                    return []
                except _CATALOG_API_NOT_FOUND_EXCEPTIONS as e:
                    # API doesn't exist - fall back to standard get_fields
                    logger.debug(
                        f"Catalog-aware get_table_req not available for fields, "
                        f"falling back: {e}"
                    )
                    fields = self.client.get_fields(db_name, table_name)
                    return [
                        {
                            "col_name": field.name,
                            "col_type": field.type,
                            "col_description": field.comment or "",
                        }
                        for field in fields
                    ]
                except TException as e:
                    # Check if it's an unsupported method error
                    error_str = str(e).lower()
                    if "unknown" in error_str or "invalid" in error_str:
                        logger.debug(
                            f"Catalog-aware API failed, falling back to get_fields: {e}"
                        )
                        fields = self.client.get_fields(db_name, table_name)
                        return [
                            {
                                "col_name": field.name,
                                "col_type": field.type,
                                "col_description": field.comment or "",
                            }
                            for field in fields
                        ]
                    raise

            # Standard HMS 2.x API
            fields = self.client.get_fields(db_name, table_name)
            return [
                {
                    "col_name": field.name,
                    "col_type": field.type,
                    "col_description": field.comment or "",
                }
                for field in fields
            ]

        return _call()

    def get_table_with_columns(
        self, db_name: str, table_name: str, catalog_name: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Get table metadata including columns in the format expected by
        hive_metastore.py's WorkUnit generation.

        Args:
            db_name: Database name
            table_name: Table name
            catalog_name: Optional catalog name for HMS 3.x
        """
        table_info = self.get_table(db_name, table_name, catalog_name)
        fields = self.get_fields(db_name, table_name, catalog_name)

        # Build column list with table-level info attached
        columns = []

        # Add regular columns
        for col in fields:
            columns.append(
                {
                    "col_name": col["col_name"],
                    "col_type": col["col_type"],
                    "col_description": col["col_description"],
                    "is_partition_col": False,
                    "table_type": table_info["table_type"],
                    "table_location": table_info["location"],
                    "create_date": str(table_info.get("create_time", "")),
                }
            )

        # Add partition columns
        for pk in table_info["partition_keys"]:
            columns.append(
                {
                    "col_name": pk["col_name"],
                    "col_type": pk["col_type"],
                    "col_description": pk["col_description"],
                    "is_partition_col": True,
                    "table_type": table_info["table_type"],
                    "table_location": table_info["location"],
                    "create_date": str(table_info.get("create_time", "")),
                }
            )

        return {
            "table_info": table_info,
            "columns": columns,
        }

    # -------------------------------------------------------------------------
    # HMS 3.x Catalog Support
    # -------------------------------------------------------------------------

    def get_catalogs(self) -> List[str]:
        """
        Get all catalog names (HMS 3.x only).

        Returns empty list if the HMS version doesn't support catalogs (HMS 2.x).
        Raises exceptions for real errors (network, auth) that shouldn't be silently ignored.
        """

        @self._get_retry_decorator()
        def _call() -> List[str]:
            try:
                response = self.client.get_catalogs()
                return list(response.names) if response.names else []
            except (AttributeError, TException) as e:
                # HMS 2.x doesn't have catalog support - this is expected
                logger.debug(f"get_catalogs not supported (likely HMS 2.x): {e}")
                return []
            # Note: TTransportException is NOT caught here - network errors should propagate

        return _call()

    # -------------------------------------------------------------------------
    # SQL Row Format Methods
    # These methods return data in the exact format expected by HiveMetastoreSource's
    # _fetch_*_rows() methods, enabling seamless integration with the parent class.
    # -------------------------------------------------------------------------

    def _iter_tables_and_views(
        self,
        databases: List[str],
        table_type_filter: Optional[Callable[[str], bool]] = None,
        catalog_name: Optional[str] = None,
    ) -> Generator[Tuple[str, str, Dict[str, Any]], None, None]:
        """
        Shared generator for iterating over tables/views.

        This DRY helper consolidates the common iteration pattern used by
        iter_table_rows(), iter_view_rows(), and iter_table_properties_rows().

        Args:
            databases: List of database names to iterate over
            table_type_filter: Optional function to filter by table_type
                               (e.g., lambda t: t != "VIRTUAL_VIEW" for tables only)
            catalog_name: Optional catalog name for HMS 3.x

        Yields:
            Tuple of (db_name, table_name, table_info dict)

        Note:
            - Table info is cached to avoid duplicate API calls when called multiple times
            - Cache keys include catalog_name to handle multi-catalog scenarios
            - Failures are tracked in dicts (not lists) to avoid duplicate entries
            - Use get_database_failures() and get_table_failures() to retrieve failures
        """
        for db_name in databases:
            # Cache key includes catalog for multi-catalog support
            db_key = DatabaseKey(catalog_name=catalog_name, db_name=db_name)

            # Skip databases that already failed
            if db_key in self._database_failures:
                continue

            try:
                table_names = self.get_all_tables(db_name, catalog_name)
            except Exception as e:
                error_msg = f"Failed to get tables for database {db_name}: {e}"
                logger.warning(error_msg)
                self._database_failures[db_key] = str(e)
                continue

            for table_name in table_names:
                # Cache key includes catalog for multi-catalog support
                table_key = TableKey(
                    catalog_name=catalog_name, db_name=db_name, table_name=table_name
                )

                # Skip tables that already failed
                if table_key in self._table_failures:
                    continue

                # Use cached table info if available
                if table_key in self._table_cache:
                    table_info = self._table_cache[table_key]
                else:
                    try:
                        table_info = self.get_table(db_name, table_name, catalog_name)
                        self._table_cache[table_key] = table_info
                    except Exception as e:
                        error_msg = f"Failed to get table {db_name}.{table_name}: {e}"
                        logger.warning(error_msg)
                        self._table_failures[table_key] = str(e)
                        continue

                # Apply table type filter
                if table_type_filter and not table_type_filter(
                    table_info["table_type"]
                ):
                    continue

                yield db_name, table_name, table_info

    @staticmethod
    def _format_create_date(create_time: Optional[int]) -> str:
        """Format Unix timestamp to YYYY-MM-DD string."""
        if not create_time:
            return ""
        try:
            dt = datetime.fromtimestamp(create_time)
            return dt.strftime("%Y-%m-%d")
        except Exception:
            return str(create_time)

    def iter_table_rows(
        self,
        databases: List[str],
        catalog_name: Optional[str] = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Iterate over table/column rows in SQL row format.

        Args:
            databases: List of database names to iterate over
            catalog_name: Optional catalog name for HMS 3.x

        Yields rows matching the format of HiveMetastoreSource._TABLES_SQL_STATEMENT:
        - schema_name: str
        - table_name: str
        - table_type: str
        - create_date: str
        - col_name: str
        - col_sort_order: int
        - col_description: str
        - col_type: str
        - is_partition_col: int (0 or 1)
        - table_location: str
        """
        for db_name, table_name, table_info in self._iter_tables_and_views(
            databases,
            table_type_filter=lambda t: t != VIRTUAL_VIEW_TYPE,  # Exclude views
            catalog_name=catalog_name,
        ):
            create_date = self._format_create_date(table_info.get("create_time"))

            # Get columns
            try:
                fields = self.get_fields(db_name, table_name, catalog_name)
            except Exception as e:
                logger.warning(f"Failed to get fields for {db_name}.{table_name}: {e}")
                table_key = TableKey(
                    catalog_name=catalog_name, db_name=db_name, table_name=table_name
                )
                self._table_failures[table_key] = f"Failed to get fields: {e}"
                fields = []

            # Yield regular columns
            for idx, col_field in enumerate(fields):
                yield {
                    "schema_name": db_name,
                    "table_name": table_name,
                    "table_type": table_info["table_type"],
                    "create_date": create_date,
                    "col_name": col_field["col_name"],
                    "col_sort_order": idx,
                    "col_description": col_field.get("col_description", ""),
                    "col_type": col_field["col_type"],
                    "is_partition_col": 0,
                    "table_location": table_info.get("location", ""),
                }

            # Yield partition columns
            for idx, pk in enumerate(table_info.get("partition_keys", [])):
                yield {
                    "schema_name": db_name,
                    "table_name": table_name,
                    "table_type": table_info["table_type"],
                    "create_date": create_date,
                    "col_name": pk["col_name"],
                    "col_sort_order": len(fields) + idx,
                    "col_description": pk.get("col_description", ""),
                    "col_type": pk["col_type"],
                    "is_partition_col": 1,
                    "table_location": table_info.get("location", ""),
                }

    def iter_view_rows(
        self,
        databases: List[str],
        catalog_name: Optional[str] = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Iterate over Hive view rows in SQL row format.

        Args:
            databases: List of database names to iterate over
            catalog_name: Optional catalog name for HMS 3.x

        Yields rows matching the format of HiveMetastoreSource._HIVE_VIEWS_SQL_STATEMENT:
        - schema_name: str
        - table_name: str
        - table_type: str (VIRTUAL_VIEW)
        - view_expanded_text: str
        - description: str
        - create_date: str
        - col_name: str
        - col_sort_order: int
        - col_description: str
        - col_type: str
        """
        for db_name, table_name, table_info in self._iter_tables_and_views(
            databases,
            table_type_filter=lambda t: t == VIRTUAL_VIEW_TYPE,  # Only views
            catalog_name=catalog_name,
        ):
            create_date = self._format_create_date(table_info.get("create_time"))
            description = table_info.get("parameters", {}).get("comment", "")

            # Get columns
            try:
                fields = self.get_fields(db_name, table_name, catalog_name)
            except Exception as e:
                logger.warning(
                    f"Failed to get fields for view {db_name}.{table_name}: {e}"
                )
                table_key = TableKey(
                    catalog_name=catalog_name, db_name=db_name, table_name=table_name
                )
                self._table_failures[table_key] = f"Failed to get view fields: {e}"
                fields = []

            for idx, col_field in enumerate(fields):
                yield {
                    "schema_name": db_name,
                    "table_name": table_name,
                    "table_type": table_info["table_type"],
                    "view_expanded_text": table_info.get("view_expanded_text", ""),
                    "description": description,
                    "create_date": create_date,
                    "col_name": col_field["col_name"],
                    "col_sort_order": idx,
                    "col_description": col_field.get("col_description", ""),
                    "col_type": col_field["col_type"],
                }

    def iter_schema_rows(
        self,
        databases: List[str],
        catalog_name: Optional[str] = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Iterate over schema/database rows in SQL row format.

        Args:
            databases: List of database names to iterate over
            catalog_name: Optional catalog name for HMS 3.x (unused, for API consistency)

        Yields rows matching the format of HiveMetastoreSource._SCHEMAS_SQL_STATEMENT:
        - schema: str
        """
        for db_name in databases:
            yield {"schema": db_name}

    def iter_table_properties_rows(
        self,
        databases: List[str],
        catalog_name: Optional[str] = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Iterate over table properties rows in SQL row format.

        Args:
            databases: List of database names to iterate over
            catalog_name: Optional catalog name for HMS 3.x

        Yields rows matching the format of HiveMetastoreSource._HIVE_PROPERTIES_SQL_STATEMENT:
        - schema_name: str
        - table_name: str
        - PARAM_KEY: str
        - PARAM_VALUE: str
        """
        for db_name, table_name, table_info in self._iter_tables_and_views(
            databases,
            table_type_filter=lambda t: t != VIRTUAL_VIEW_TYPE,  # Exclude views
            catalog_name=catalog_name,
        ):
            # Yield each parameter as a row
            for param_key, value in table_info.get("parameters", {}).items():
                if value is not None:
                    yield {
                        "schema_name": db_name,
                        "table_name": table_name,
                        "PARAM_KEY": param_key,
                        "PARAM_VALUE": str(value),
                    }



from __future__ import annotations

import warnings
from datetime import datetime
from typing import Dict, Optional, Type, Union

from typing_extensions import Self

import datahub.metadata.schema_classes as models
from datahub.cli.cli_utils import first_non_null
from datahub.emitter.mce_builder import DEFAULT_ENV
from datahub.errors import (
    IngestionAttributionWarning,
)
from datahub.metadata.urns import DataFlowUrn, Urn
from datahub.sdk._attribution import is_ingestion_attribution
from datahub.sdk._shared import (
    DomainInputType,
    HasContainer,
    HasDomain,
    HasInstitutionalMemory,
    HasOwnership,
    HasPlatformInstance,
    HasStructuredProperties,
    HasSubtype,
    HasTags,
    HasTerms,
    LinksInputType,
    OwnersInputType,
    ParentContainerInputType,
    StructuredPropertyInputType,
    TagsInputType,
    TermsInputType,
    make_time_stamp,
    parse_time_stamp,
)
from datahub.sdk.entity import Entity, ExtraAspectsType
from datahub.utilities.sentinels import Unset, unset


class DataFlow(
    HasPlatformInstance,
    HasSubtype,
    HasOwnership,
    HasContainer,
    HasInstitutionalMemory,
    HasTags,
    HasTerms,
    HasDomain,
    HasStructuredProperties,
    Entity,
):
    """Represents a dataflow in DataHub.
    A dataflow represents a collection of data, such as a table, view, or file.
    This class provides methods for managing dataflow metadata including schema,
    lineage, and various aspects like ownership, tags, and terms.
    """

    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[DataFlowUrn]:
        """Get the URN type for dataflows.
        Returns:
            The DataflowUrn class.
        """
        return DataFlowUrn

    def __init__(
        self,
        *,
        # Identity.
        name: str,
        platform: str,
        display_name: Optional[str] = None,
        platform_instance: Optional[str] = None,
        env: str = DEFAULT_ENV,
        # Dataflow properties.
        description: Optional[str] = None,
        external_url: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
        created: Optional[datetime] = None,
        last_modified: Optional[datetime] = None,
        # Standard aspects.
        subtype: Optional[str] = None,
        owners: Optional[OwnersInputType] = None,
        links: Optional[LinksInputType] = None,
        tags: Optional[TagsInputType] = None,
        terms: Optional[TermsInputType] = None,
        domain: Optional[DomainInputType] = None,
        parent_container: ParentContainerInputType | Unset = unset,
        structured_properties: Optional[StructuredPropertyInputType] = None,
        extra_aspects: ExtraAspectsType = None,
    ):
        """Initialize a new Dataflow instance.
        Args:
            platform: The platform this dataflow belongs to (e.g. "mysql", "snowflake").
            name: The name of the dataflow.
            platform_instance: Optional platform instance identifier.
            env: The environment this dataflow belongs to (default: DEFAULT_ENV).
            description: Optional description of the dataflow.
            display_name: Optional display name for the dataflow.
            external_url: Optional URL to external documentation or source.
            custom_properties: Optional dictionary of custom properties.
            created: Optional creation timestamp.
            last_modified: Optional last modification timestamp.
            subtype: Optional subtype of the dataflow.
            owners: Optional list of owners.
            links: Optional list of links.
            tags: Optional list of tags.
            terms: Optional list of glossary terms.
            domain: Optional domain this dataflow belongs to.
            extra_aspects: Optional list of additional aspects.
            upstreams: Optional upstream lineage information.
        """
        urn = DataFlowUrn.create_from_ids(
            orchestrator=platform,
            flow_id=name,
            env=env,
            platform_instance=platform_instance,
        )
        super().__init__(urn)
        self._set_extra_aspects(extra_aspects)

        self._set_platform_instance(urn.orchestrator, platform_instance)

        # Initialize DataFlowInfoClass directly with name
        self._setdefault_aspect(models.DataFlowInfoClass(name=display_name or name))
        self._ensure_dataflow_props().env = env

        if description is not None:
            self.set_description(description)
        if display_name is not None:
            self.set_display_name(display_name)
        if external_url is not None:
            self.set_external_url(external_url)
        if custom_properties is not None:
            self.set_custom_properties(custom_properties)
        if created is not None:
            self.set_created(created)
        if last_modified is not None:
            self.set_last_modified(last_modified)
        if subtype is not None:
            self.set_subtype(subtype)
        if owners is not None:
            self.set_owners(owners)
        if links is not None:
            self.set_links(links)
        if tags is not None:
            self.set_tags(tags)
        if terms is not None:
            self.set_terms(terms)
        if domain is not None:
            self.set_domain(domain)
        if parent_container is not unset:
            self._set_container(parent_container)
        if structured_properties is not None:
            for key, value in structured_properties.items():
                self.set_structured_property(property_urn=key, values=value)

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, DataFlowUrn)
        entity = cls(
            platform=urn.orchestrator,
            name=urn.flow_id,
        )
        return entity._init_from_graph(current_aspects)

    @property
    def urn(self) -> DataFlowUrn:
        return self._urn  # type: ignore

    def _ensure_dataflow_props(self) -> models.DataFlowInfoClass:
        props = self._get_aspect(models.DataFlowInfoClass)
        if props is None:
            # Use name from URN as fallback
            props = models.DataFlowInfoClass(name=self.urn.flow_id)
            self._set_aspect(props)
        return props

    def _get_editable_props(self) -> Optional[models.EditableDataFlowPropertiesClass]:
        return self._get_aspect(models.EditableDataFlowPropertiesClass)

    def _ensure_editable_props(self) -> models.EditableDataFlowPropertiesClass:
        # Note that most of the fields in this aspect are not used.
        # The only one that's relevant for us is the description.
        return self._setdefault_aspect(models.EditableDataFlowPropertiesClass())

    @property
    def description(self) -> Optional[str]:
        """Get the description of the dataflow.
        Returns:
            The description if set, None otherwise.
        """
        editable_props = self._get_editable_props()
        return first_non_null(
            [
                editable_props.description if editable_props is not None else None,
                self._ensure_dataflow_props().description,
            ]
        )

    def set_description(self, description: str) -> None:
        """Set the description of the dataflow.
        Args:
            description: The description to set.
        Note:
            If called during ingestion, this will warn if overwriting
            a non-ingestion description.
        """
        if is_ingestion_attribution():
            editable_props = self._get_editable_props()
            if editable_props is not None and editable_props.description is not None:
                warnings.warn(
                    "Overwriting non-ingestion description from ingestion is an anti-pattern.",
                    category=IngestionAttributionWarning,
                    stacklevel=2,
                )
                # Force the ingestion description to show up.
                editable_props.description = None

            self._ensure_dataflow_props().description = description
        else:
            self._ensure_editable_props().description = description

    @property
    def name(self) -> str:
        """Get the name of the dataflow.
        Returns:
            The name of the dataflow.
        """
        return self.urn.flow_id

    @property
    def display_name(self) -> Optional[str]:
        """Get the display name of the dataflow.
        Returns:
            The display name if set, None otherwise.
        """
        return self._ensure_dataflow_props().name

    def set_display_name(self, display_name: str) -> None:
        """Set the display name of the dataflow.
        Args:
            display_name: The display name to set.
        """
        self._ensure_dataflow_props().name = display_name

    @property
    def external_url(self) -> Optional[str]:
        """Get the external URL of the dataflow.
        Returns:
            The external URL if set, None otherwise.
        """
        return self._ensure_dataflow_props().externalUrl

    def set_external_url(self, external_url: str) -> None:
        """Set the external URL of the dataflow.
        Args:
            external_url: The external URL to set.
        """
        self._ensure_dataflow_props().externalUrl = external_url

    @property
    def custom_properties(self) -> Dict[str, str]:
        """Get the custom properties of the dataflow.
        Returns:
            Dictionary of custom properties.
        """
        return self._ensure_dataflow_props().customProperties

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        """Set the custom properties of the dataflow.
        Args:
            custom_properties: Dictionary of custom properties to set.
        """
        self._ensure_dataflow_props().customProperties = custom_properties

    @property
    def created(self) -> Optional[datetime]:
        """Get the creation timestamp of the dataflow.
        Returns:
            The creation timestamp if set, None otherwise.
        """
        return parse_time_stamp(self._ensure_dataflow_props().created)

    def set_created(self, created: datetime) -> None:
        """Set the creation timestamp of the dataflow.
        Args:
            created: The creation timestamp to set.
        """
        self._ensure_dataflow_props().created = make_time_stamp(created)

    @property
    def last_modified(self) -> Optional[datetime]:
        """Get the last modification timestamp of the dataflow.
        Returns:
            The last modification timestamp if set, None otherwise.
        """
        return parse_time_stamp(self._ensure_dataflow_props().lastModified)

    def set_last_modified(self, last_modified: datetime) -> None:
        self._ensure_dataflow_props().lastModified = make_time_stamp(last_modified)

    @property
    def env(self) -> Optional[Union[str, models.FabricTypeClass]]:
        """Get the environment of the dataflow."""
        return self._ensure_dataflow_props().env



import logging
from typing import Any, Dict, List, Optional

from gql import gql

from datahub.api.graphql.base import BaseApi

logger = logging.getLogger(__name__)


class Assertion(BaseApi):
    ASSERTION_QUERY = """
query dataset($urn: String!, $start: Int, $count: Int, $status: AssertionRunStatus,$limit: Int, $startTimeMillis:Long, $endTimeMillis:Long, $filter:FilterInput) {
  dataset(urn: $urn) {
    assertions(start: $start, count: $count){
      __typename
      total
      assertions{
        __typename
        runEvents(status: $status, limit: $limit, startTimeMillis: $startTimeMillis, endTimeMillis: $endTimeMillis, filter: $filter) {
          total
          failed
          succeeded
          runEvents {
            __typename
            timestampMillis
            partitionSpec {
               __typename
              type
              partition
              timePartition {
                startTimeMillis
                durationMillis
              }
            }
            result {
              __typename
              type
              rowCount
              missingCount
              unexpectedCount
              actualAggValue
              externalUrl
            }
            assertionUrn
          }
        }
      }
    }
  }
}
"""

    def query_assertion(
        self,
        urn: str,
        status: Optional[str] = None,
        start_time_millis: Optional[int] = None,
        end_time_millis: Optional[int] = None,
        limit: Optional[int] = None,
        filter: Optional[Dict[str, Optional[str]]] = None,
    ) -> List[Dict[Any, Any]]:
        r"""
        Query assertions for a dataset.

        :param urn: The DataHub dataset unique identifier.
        :param status: The assertion status to filter for. Every status will be accepted if it is not set.
            See valid status at https://docs.datahub.com/docs/graphql/enums#assertionrunstatus
        :param start_time_millis: The start time in milliseconds from the assertions will be queried.
        :param end_time_millis: The end time in milliseconds until the assertions will be queried.
        :param filter: Additional key value filters which will be applied as AND query
        """

        result = self.client.execute(
            gql(Assertion.ASSERTION_QUERY),
            variable_values={
                "urn": urn,
                "filter": self.gen_filter(filter) if filter else None,
                "limit": limit,
                "status": status,
                "startTimeMillis": start_time_millis,
                "endTimeMillis": end_time_millis,
            },
        )

        assertions = []
        if "dataset" in result and "assertions" in result["dataset"]:
            assertions = result["dataset"]["assertions"]["assertions"]

        return assertions
